import base64
import json
import re
from datetime import datetime
from StringIO import StringIO

import OFS
from BTrees.OOBTree import OOBTree
from DateTime import DateTime
from Persistence.mapping import PersistentMapping as PM1
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping as PM2
from plone.app.blob.field import BlobWrapper
from plone.app.blob.utils import openBlob
from plone.app.contentlisting.contentlisting import ContentListing
from plone.app.textfield.value import RichTextValue
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedFile
from Products.ZCatalog.Lazy import LazyCat
from ZODB.blob import Blob
from zope.dottedname.resolve import resolve
from ZPublisher.HTTPRequest import record

_filedata_marker = 'filedata://'
_deferred_marker = 'deferred://'
_type_marker = 'type://'
_uid_marker = 'uid://'
_uid_separator = '||||'
_date_re = re.compile('^[0-9]{4}\-[0-9]{2}\-[0-9]{2}.*$')


class BaseTypeSerializer(object):
    klass = None
    toklass = None

    @classmethod
    def get_type_name(kls):
        return "%s.%s" % (kls.klass.__module__, kls.klass.__name__)

    @classmethod
    def serialize(kls, obj):
        if hasattr(obj, 'aq_base'):
            obj = obj.aq_base
        data = kls._serialize(obj)
        results = {
            'type': kls.get_type_name(),
            'data': data
        }
        return _type_marker + dumps(results)

    @classmethod
    def _serialize(kls, obj):
        return kls.toklass(obj)

    @classmethod
    def deserialize(kls, data):
        return kls._deserialize(data)

    @classmethod
    def _deserialize(kls, data):
        return kls.klass(data)


class PM1Serializer(BaseTypeSerializer):
    klass = PM1
    toklass = dict


class PM2Serializer(PM1Serializer):
    klass = PM2


class PersistentDictSerializer(BaseTypeSerializer):
    klass = PersistentDict
    toklass = dict


class OOBTreeSerializer(BaseTypeSerializer):
    klass = OOBTree
    toklass = dict


class PersistentListSerializer(BaseTypeSerializer):
    klass = PersistentList
    toklass = list


class setSerializer(BaseTypeSerializer):
    klass = set
    toklass = list


class OFSFileSerializer(BaseTypeSerializer):
    klass = OFS.Image.File

    @classmethod
    def _serialize(kls, obj):
        try:
            data = str(obj.data)
        except Exception:
            data = str(obj.data.data)
        return {
            'data': base64.b64encode(data),
            'id': obj.id(),
            'title': obj.title,
            'content_type': obj.content_type
        }

    @classmethod
    def _deserialize(kls, data):
        file = base64.b64decode(data['data'])
        id = data['id']
        title = data['title']
        ct = data['content_type']
        return kls.klass(id, title, file, ct)


class BlobSerializer(BaseTypeSerializer):
    klass = Blob

    @classmethod
    def _serialize(kls, obj):
        blobfi = openBlob(obj)
        data = blobfi.read()
        blobfi.close()
        return {'data': base64.b64encode(data)}

    @classmethod
    def _deserialize(kls, data):
        blob = Blob()
        bfile = blob.open('w')
        data = base64.b64decode(data['data'])
        bfile.write(data)
        bfile.close()
        return blob


class OFSImageSerializer(OFSFileSerializer):
    klass = OFS.Image.Image


class DateTimeSerializer(BaseTypeSerializer):
    klass = DateTime

    @classmethod
    def get_type_name(kls):
        return 'DateTime.DateTime'

    @classmethod
    def _serialize(kls, obj):
        return obj.ISO8601()

    @classmethod
    def _deserialize(kls, data):
        return DateTime(data)


class datetimeSerializer(BaseTypeSerializer):
    klass = datetime

    @classmethod
    def _serialize(kls, obj):
        return obj.isoformat()

    @classmethod
    def _deserialize(kls, data):
        try:
            return datetime.strptime(data, '%Y-%m-%dT%H:%M:%S.%f')
        except Exception:
            return datetime.strptime(data, '%Y-%m-%dT%H:%M:%S')


class recordSerializer(BaseTypeSerializer):
    klass = record
    toklass = dict

    @classmethod
    def _deserialize(kls, data):
        rec = record()
        for key, value in data.items():
            setattr(rec, key, value)
        return rec


class ContentListingSerializer(BaseTypeSerializer):
    klass = ContentListing
    toklass = list

    @classmethod
    def _serialize(kls, obj):
        return list(obj._basesequence)

    @classmethod
    def _deserialize(kls, data):
        return LazyCat(data)


class BlobWrapperSerializer(BaseTypeSerializer):
    klass = BlobWrapper

    @classmethod
    def _serialize(kls, obj):
        blob = obj.getBlob()
        blobfi = openBlob(blob)
        data = blobfi.read()
        blobfi.close()
        return {
            'data': base64.b64encode(data),
            'filename': obj.getFilename()}

    @classmethod
    def _deserialize(kls, data):
        io = StringIO(base64.b64decode(data['data']))
        io.filename = data['filename']
        return io


class NamedBlobFileSerializer(BaseTypeSerializer):
    klass = NamedBlobFile

    @classmethod
    def _serialize(cls, obj):
        return {
            'data': base64.b64encode(obj.data),
            'filename': obj.filename,
            'content_type': obj.contentType
        }

    @classmethod
    def _deserialize(cls, data):
        return NamedBlobFile(
            base64.b64decode(data['data']),
            filename=data['filename'],
            contentType=data['content_type'].encode('utf-8')
        )


class NamedBlobImageSerializer(NamedBlobFileSerializer):
    klass = NamedBlobImage

    @classmethod
    def _deserialize(cls, data):
        return NamedBlobImage(
            base64.b64decode(data['data']),
            filename=data['filename'],
            contentType=data['content_type'].encode('utf-8')
        )


class RichTextValueSerializer(BaseTypeSerializer):
    klass = RichTextValue

    @classmethod
    def _serialize(cls, obj):
        return {
            'raw': obj.raw,
            'mimeType': obj.mimeType,
            'outputMimeType': obj.outputMimeType,
            'encoding': obj.encoding
        }

    @classmethod
    def _deserialize(cls, data):
        return RichTextValue(
            raw=data['raw'],
            mimeType=data['mimeType'].encode('utf-8'),
            outputMimeType=data['outputMimeType'].encode('utf-8'),
            encoding=data['encoding'].encode('utf-8'))


_serializers = {
    PM1: PM1Serializer,
    PM2: PM2Serializer,
    PersistentDict: PersistentDictSerializer,
    OOBTree: OOBTreeSerializer,
    PersistentList: PersistentListSerializer,
    set: setSerializer,
    OFS.Image.Image: OFSImageSerializer,
    OFS.Image.File: OFSFileSerializer,
    DateTime: DateTimeSerializer,
    datetime: datetimeSerializer,
    record: recordSerializer,
    Blob: BlobSerializer,
    ContentListing: ContentListingSerializer,
    BlobWrapper: BlobWrapperSerializer,
    NamedBlobImage: NamedBlobImageSerializer,
    NamedBlobFile: NamedBlobFileSerializer,
    NamedFile: NamedBlobFileSerializer,
    RichTextValue: RichTextValueSerializer
}


class Deferred:
    pass


def custom_handler(obj):
    if hasattr(obj, 'aq_base'):
        obj = obj.aq_base
    _type = type(obj)
    if _type.__name__ == 'instance':
        _type = obj.__class__
    if _type in _serializers:
        serializer = _serializers[_type]
        return serializer.serialize(obj)
    else:
        return None
    return obj


def custom_decoder(d):
    if isinstance(d, list):
        pairs = enumerate(d)
    elif isinstance(d, dict):
        pairs = d.items()
    result = []
    for k, v in pairs:
        if isinstance(v, basestring):
            if v.startswith(_filedata_marker):
                if v == _filedata_marker + _deferred_marker:
                    v = Deferred
                else:
                    filedata = v[len(_filedata_marker):]
                    if filedata:
                        v = StringIO(base64.b64decode(filedata))
                    else:
                        v = ''
            elif v.startswith(_type_marker):
                v = v[len(_type_marker):]
                results = loads(v)
                try:
                    _type = resolve(results['type'])
                    serializer = _serializers[_type]
                    v = serializer.deserialize(results['data'])
                except ImportError:
                    pass
        elif isinstance(v, (dict, list)):
            v = custom_decoder(v)
        result.append((k, v))
    if isinstance(d, list):
        return [x[1] for x in result]
    elif isinstance(d, dict):
        return dict(result)


def loads(data):
    return json.loads(data, object_hook=custom_decoder)


def dumps(data):
    return json.dumps(data, default=custom_handler)



# -*- coding: utf-8 -*-
import unittest

from castle.cms.browser.site.sitemap import SiteMapView
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from plone import api
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles


class TestSiteMap(unittest.TestCase):

    layer = CASTLE_PLONE_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']

        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        if 'front-page' not in self.portal:
            self.page = api.content.create(type='Document', id='front-page',
                                           container=self.portal)
            self.portal.setDefaultPage('front-page')

    def test_get_objects(self):
        view = SiteMapView(self.portal, self.request)
        paths = [o for o in view.objects()]
        self.assertEquals(paths[0]['loc'], self.portal.absolute_url())

    def test_no_private_content(self):
        folder = api.content.create(
            type='Folder',
            id='test-folder',
            container=self.portal
        )
        page = api.content.create(
            type='Document',
            id='test-page',
            container=folder
        )
        api.content.transition(obj=page, to_state='published')
        view = SiteMapView(self.portal, self.request)
        locations = [object['loc'] for object in view.objects()]
        self.assertFalse('http://nohost/plone/test-folder' in locations)
        self.assertFalse('http://nohost/plone/test-folder/test-page' in locations)
        api.content.transition(obj=folder, to_state='published')
        locations = [object['loc'] for object in view.objects()]
        self.assertTrue('http://nohost/plone/test-folder' in locations)
        self.assertTrue('http://nohost/plone/test-folder/test-page' in locations)



import logging

from plone import api
from plone.app.redirector.browser import FourOhFourView
from plone.app.redirector.interfaces import IRedirectionStorage
from zExceptions import Redirect
from zope.component import queryUtility
from six.moves import urllib
from six.moves.urllib.parse import quote
from six.moves.urllib.parse import unquote

from castle.cms import archival
from castle.cms import shield
from castle.cms.files import aws


logger = logging.getLogger('castle.cms')


class FourOhFour(FourOhFourView):

    def __call__(self):
        shield.protect(self.request, recheck=True)
        self.notfound = self.context
        self.context = api.portal.get()
        if '++' in self.request.URL:
            self.request.response.setStatus(404)
            try:
                return self.index()
            except Exception:
                logger.warn("Failed to render 404 template, had to return simple response")
                return "not found"

        archive_storage = archival.Storage(self.context)
        site_url = self.context.absolute_url()
        path = self.request.ACTUAL_URL[len(site_url):].rstrip('/')

        wants_view = False
        if path.endswith('/view'):
            wants_view = True
            path = path.rsplit('/view', 1)[0]

        new_url = None
        if path.startswith('/resolveuid'):
            uid = path.replace('/resolveuid/', '')
            try:
                new_url = archive_storage.get_archive_url_by_uid(uid)
            except Exception:
                pass
        else:
            try:
                new_url = archive_storage.get_archive_url_by_path(path, wants_view)
            except Exception:
                pass
        if new_url:
            # XXX need to force redirect this way since normal redirect
            # gets overridden with 404
            if self.request.environ.get('QUERY_STRING'):
                new_url += '?' + self.request.environ['QUERY_STRING']
            raise Redirect(aws.swap_url(new_url))

        self.attempt_redirect()

        self.request.response.setStatus(404)
        return self.index()

    def attempt_redirect(self):
        url = self._url()
        if not url:
            return False

        try:
            old_path_elements = self.request.physicalPathFromURL(url)
        except ValueError:
            return False

        storage = queryUtility(IRedirectionStorage)
        if storage is None:
            return False

        old_path = '/'.join(old_path_elements)

        # First lets try with query string in cases or content migration

        new_path = None

        query_string = self.request.QUERY_STRING
        if query_string:
            new_path = storage.get("%s?%s" % (old_path, query_string))
            # if we matched on the query_string we don't want to include it
            # in redirect
            if new_path:
                query_string = ''

        if not new_path:
            new_path = storage.get(old_path)

        if not new_path:
            new_path = self.find_redirect_if_view(old_path_elements, storage)

        if not new_path:
            new_path = self.find_redirect_if_template(
                url,
                old_path_elements,
                storage)

        if not new_path:
            return False

        url = urllib.parse.urlsplit(new_path)
        if url.netloc:
            # External URL
            # avoid double quoting
            url_path = unquote(url.path)
            url_path = quote(url_path)
            url = urllib.parse.SplitResult(
                *(url[:2] + (url_path, ) + url[3:])).geturl()
        else:
            url = self.request.physicalPathToURL(new_path)

        # some analytics programs might use this info to track
        if query_string:
            url += "?" + query_string
        raise Redirect(url)
        return True



from castle.cms.interfaces import ISecuritySchema
from castle.cms.widgets import SelectFieldWidget
from plone.app.registry.browser import controlpanel
from Products.CMFPlone.controlpanel.browser import security
from z3c.form.field import Fields
from zope.schema import Int
from plone import api


class SecurityControlPanelForm(security.SecurityControlPanelForm):
    schema = ISecuritySchema

    def updateFields(self):
        super(SecurityControlPanelForm, self).updateFields()
        self.fields['restrict_logins_to_countries'].widgetFactory = SelectFieldWidget
        timeout = api.portal.get().acl_users.session.timeout
        timeout_field = Int(__name__='session_timeout',
                            title=u'Session Timeout',
                            description=u'Duration of user session timeout '
                                        u'(seconds)',
                            required=False,
                            default=timeout)
        self.fields += Fields(timeout_field)

    def applyChanges(self, data):
        super(SecurityControlPanelForm, self).applyChanges(data)
        timeout = int(data['session_timeout'])
        self.context.acl_users.session.timeout = timeout


class SecurityControlPanel(controlpanel.ControlPanelFormWrapper):
    form = SecurityControlPanelForm



import argparse
import base64
import errno
import logging
import time
import os
import re
from datetime import datetime
from fnmatch import fnmatch
from StringIO import StringIO

import OFS
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManager import setSecurityPolicy
from Acquisition import ImplicitAcquisitionWrapper
from Acquisition import aq_parent
from BTrees.OOBTree import OOBTree
from DateTime import DateTime
from lxml.html import fromstring
from OFS.interfaces import IFolder
from Persistence.mapping import PersistentMapping as PM1  # noqa
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping as PM2  # noqa
from plone.app.blob.field import BlobWrapper
from plone.app.blob.utils import openBlob
from Products.Archetypes import Field
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.tests.base.security import OmnipotentUser
from Products.CMFCore.tests.base.security import PermissiveSecurityPolicy
from Products.ZCatalog.Lazy import LazyCat
from Testing.makerequest import makerequest
from ZODB.blob import Blob
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility
from zope.component.hooks import setSite
from zope.interface import Interface
from zope.schema import getFieldsInOrder
from ZPublisher.HTTPRequest import record


logger = logging.getLogger(__name__)

try:
    from plone.uuid.interfaces import IUUID
except ImportError:
    def IUUID(obj, default=None):
        return default

try:
    from Products.PressRoom.content.PressContact import PressContact
except ImportError:
    PressContact = False

try:
    from Products.CMFPlone import defaultpage
except ImportError:
    defaultpage = None

try:
    import json
except ImportError:
    import simplejson as json

try:
    from plone.app.contentlisting.contentlisting import ContentListing
except ImportError:
    ContentListing = object()

try:
    from plone.dexterity.interfaces import IDexterityContent
except ImportError:
    class IDexterityContent(Interface):
        pass

try:
    from plone.app.blocks.layoutbehavior import ILayoutAware
except ImportError:
    class ILayoutAware(Interface):
        pass


try:
    from plone.namedfile.file import NamedBlobFile
except ImportError:
    print('unable to import NamedBlobFile')
    NamedBlobFile = object()

try:
    from plone.namedfile.file import NamedFile
except ImportError:
    print('unabled to import NamedFile')
    NamedFile = object()


try:
    from plone.namedfile.file import NamedBlobImage
except ImportError:
    print('unable to import NamedBlobImage')
    NamedBlobImage = object()

try:
    from plone.app.textfield import RichText
    from plone.app.textfield.value import RichTextValue
except ImportError:
    RichText = None
    RichTextValue = None


parser = argparse.ArgumentParser(
    description='...')
parser.add_argument('--overwrite', dest='overwrite', default=False)
parser.add_argument('--admin-user', dest='admin_user', default='admin')
parser.add_argument('--site-id', dest='site_id', default='Plone')
parser.add_argument('--dir', dest='dir', default='./export')
parser.add_argument('--interval', dest='interval', default=0)
parser.add_argument('--modifiedsince', dest='modifiedsince')
parser.add_argument('--createdsince', dest='createdsince')
parser.add_argument(
    '--path-filter', dest='path_filter',
    default=None, required=False)
args, _ = parser.parse_known_args()


def spoof_request(app):
    """
    Make REQUEST variable to be available on the Zope application server.

    This allows acquisition to work properly
    """
    _policy = PermissiveSecurityPolicy()
    _oldpolicy = setSecurityPolicy(_policy)  # noqa
    newSecurityManager(None, OmnipotentUser().__of__(app.acl_users))
    return makerequest(app)

app = spoof_request(app)  # noqa

user = app.acl_users.getUser(args.admin_user)  # noqa
try:
    newSecurityManager(None, user.__of__(app.acl_users))  # noqa
except Exception:
    logger.error('Unknown admin user; '
                 'specify an existing Zope admin user with --admin-user (default is admin)')  # noqa
    exit(-1)
site = app[args.site_id]
setSite(site)

_filedata_marker = 'filedata://'
_deferred_marker = 'deferred://'
_type_marker = 'type://'
_uid_marker = 'uid://'
_uid_separator = '||||'
_date_re = re.compile('^[0-9]{4}\-[0-9]{2}\-[0-9]{2}.*$')


export_folder = os.path.abspath(args.dir)
catalog = site.portal_catalog
workflow = site.portal_workflow
ptool = site.plone_utils
site_path = '/'.join(site.getPhysicalPath())


class BaseTypeSerializer(object):
    klass = None
    toklass = None

    @classmethod
    def get_type_name(cls):
        return "%s.%s" % (cls.klass.__module__, cls.klass.__name__)

    @classmethod
    def serialize(cls, obj):
        if hasattr(obj, 'aq_base'):
            obj = obj.aq_base
        data = cls._serialize(obj)
        results = {
            'type': cls.get_type_name(),
            'data': data
        }
        return _type_marker + dumps(results)

    @classmethod
    def _serialize(cls, obj):
        return cls.toklass(obj)

    @classmethod
    def deserialize(cls, data):
        return cls._deserialize(data)

    @classmethod
    def _deserialize(cls, data):
        return cls.klass(data)


class PM1Serializer(BaseTypeSerializer):
    klass = PM1
    toklass = dict


class PM2Serializer(PM1Serializer):
    klass = PM2


class PersistentDictSerializer(BaseTypeSerializer):
    klass = PersistentDict
    toklass = dict


class OOBTreeSerializer(BaseTypeSerializer):
    klass = OOBTree
    toklass = dict


class PersistentListSerializer(BaseTypeSerializer):
    klass = PersistentList
    toklass = list


class setSerializer(BaseTypeSerializer):
    klass = set
    toklass = list


class OFSFileSerializer(BaseTypeSerializer):
    klass = OFS.Image.File

    @classmethod
    def _serialize(cls, obj):
        try:
            data = str(obj.data)
        except Exception:
            print('Error in OFSFileSerializer while serializing {}'.format(
                '/'.join(obj.getPhysicalPath())))
            data = str(obj.data.data)
        return {
            'data': base64.b64encode(data),
            'id': obj.id(),
            'title': obj.title,
            'content_type': obj.content_type
        }

    @classmethod
    def _deserialize(cls, data):
        file = base64.b64decode(data['data'])
        id = data['id']
        title = data['title']
        ct = data['content_type']
        return cls.klass(id, title, file, ct)


class BlobSerializer(BaseTypeSerializer):
    klass = Blob

    @classmethod
    def _serialize(cls, obj):
        blobfi = openBlob(obj)
        data = blobfi.read()
        blobfi.close()
        return {'data': base64.b64encode(data)}

    @classmethod
    def _deserialize(cls, data):
        blob = Blob()
        bfile = blob.open('w')
        data = base64.b64decode(data['data'])
        bfile.write(data)
        bfile.close()
        return blob


class OFSImageSerializer(OFSFileSerializer):
    klass = OFS.Image.Image


class DateTimeSerializer(BaseTypeSerializer):
    klass = DateTime

    @classmethod
    def get_type_name(cls):
        return 'DateTime.DateTime'

    @classmethod
    def _serialize(cls, obj):
        return obj.ISO8601()

    @classmethod
    def _deserialize(cls, data):
        return DateTime(data)


class datetimeSerializer(BaseTypeSerializer):
    klass = datetime

    @classmethod
    def _serialize(cls, obj):
        return obj.isoformat()

    @classmethod
    def _deserialize(cls, data):
        return datetime.strptime(data, '%Y-%m-%dT%H:%M:%S.%f')


class recordSerializer(BaseTypeSerializer):
    klass = record
    toklass = dict

    @classmethod
    def _deserialize(cls, data):
        rec = record()
        for key, value in data.items():
            setattr(rec, key, value)
        return rec


class ContentListingSerializer(BaseTypeSerializer):
    klass = ContentListing
    toklass = list

    @classmethod
    def _serialize(cls, obj):
        return list(obj._basesequence)

    @classmethod
    def _deserialize(cls, data):
        return LazyCat(data)


class BlobWrapperSerializer(BaseTypeSerializer):
    klass = BlobWrapper

    @classmethod
    def _serialize(cls, obj):
        blob = obj.getBlob()
        blobfi = openBlob(blob)
        data = blobfi.read()
        blobfi.close()
        return {
            'data': base64.b64encode(data),
            'filename': obj.getFilename()}

    @classmethod
    def _deserialize(cls, data):
        io = StringIO(base64.b64decode(data['data']))
        io.filename = data['filename']
        return io


class NamedBlobFileSerializer(BaseTypeSerializer):
    klass = NamedBlobFile

    @classmethod
    def _serialize(cls, obj):
        return {
            'data': base64.b64encode(obj.data),
            'filename': obj.filename,
            'content_type': obj.contentType
        }

    @classmethod
    def _deserialize(cls, data):
        realdata = data['data']
        if len(realdata) < 10:  # arbitrary
            print(
                "short data found in NamedBlobFileSerializer _deserialize")
        return NamedBlobFile(
            base64.b64decode(data['data']),
            filename=data['filename'],
            contentType=data['content_type'].encode('utf-8')
        )


class NamedBlobImageSerializer(NamedBlobFileSerializer):
    klass = NamedBlobImage

    @classmethod
    def _deserialize(cls, data):
        realdata = data['data']
        if len(realdata) < 10:  # arbitrary
            print(
                "short data found in NamedBlobImageSerializer _deserialize")
        return NamedBlobImage(
            base64.b64decode(realdata),
            filename=data['filename'],
            contentType=data['content_type']
        )


class RichTextValueSerializer(BaseTypeSerializer):
    klass = RichTextValue

    @classmethod
    def _serialize(cls, obj):
        return {
            'raw': obj.raw,
            'mimeType': obj.mimeType,
            'outputMimeType': obj.outputMimeType,
            'encoding': obj.encoding
        }

    @classmethod
    def _deserialize(cls, data):
        return RichTextValue(
            raw=data['raw'],
            mimeType=data['mimeType'],
            outputMimeType=data['outputMimeType'],
            encoding=data['encoding'])


_serializers = {
    PM1: PM1Serializer,
    PM2: PM2Serializer,
    PersistentDict: PersistentDictSerializer,
    OOBTree: OOBTreeSerializer,
    PersistentList: PersistentListSerializer,
    set: setSerializer,
    OFS.Image.Image: OFSImageSerializer,
    Field.Image: OFSImageSerializer,
    OFS.Image.File: OFSFileSerializer,
    DateTime: DateTimeSerializer,
    datetime: datetimeSerializer,
    record: recordSerializer,
    Blob: BlobSerializer,
    ContentListing: ContentListingSerializer,
    BlobWrapper: BlobWrapperSerializer,
    NamedFile: NamedBlobFileSerializer,
    NamedBlobFile: NamedBlobFileSerializer,
    NamedBlobImage: NamedBlobImageSerializer
}


if PressContact:
    class ContentObjectSerializer(BaseTypeSerializer):
        klass = PressContact

        @classmethod
        def _serialize(cls, obj):
            return _uid_marker + obj.UID()

        @classmethod
        def _deserialize(cls, data):
            return None
    _serializers[PressContact] = ContentObjectSerializer


if RichTextValue is not None:
    _serializers[RichTextValue] = RichTextValueSerializer


class Deferred:
    pass


def custom_handler(obj):
    if hasattr(obj, 'aq_base'):
        obj = obj.aq_base
    _type = type(obj)
    if _type.__name__ == 'instance':
        _type = obj.__class__
    if _type in _serializers:
        serializer = _serializers[_type]
        return serializer.serialize(obj)
    else:
        print('NOT SERIALIZING {}'.format(obj))
        return None
    return obj


def dumps(data):
    return json.dumps(data, default=custom_handler)


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise


def get_uid(obj):
    value = IUUID(obj, None)
    if not value and hasattr(obj, 'UID'):
        value = obj.UID()
    return value


def create_path(path):
    path = path[len(site_path):]
    xpath = os.path.join(export_folder, path.lstrip('/'))
    xpathsplit = xpath.split('/')
    xfolderpath = '/'.join(xpathsplit[:-1])
    mkdir_p(xfolderpath)
    return xpath


class ContentExporter(object):

    im_width = 0
    image_to_lead = None

    def __init__(self, obj):
        self.obj = obj

    def get_field_data(self):
        pass

    def get_referenced_images(self, data):
        images = []
        for key, value in data.items():
            if not isinstance(value, basestring):
                continue
            try:
                dom = fromstring(value)
            except Exception:
                continue
            for el in dom.cssselect('img'):
                src = el.attrib.get('src', '')
                if 'resolveuid' not in src:
                    continue
                uid = src.split('resolveuid/')[-1].split(
                    '/@@images')[0].split('/image_')[0]
                brains = catalog(UID=uid)
                if len(brains) > 0:
                    images.append(brains[0].getObject())
        return images

    def get_object_data(self):
        is_dp = False
        if not IFolder.providedBy(self.obj):
            if defaultpage:
                container = aq_parent(self.obj)
                if container:
                    is_dp = defaultpage.is_default_page(container, self.obj)
            else:
                is_dp = ptool.isDefaultPage(self.obj)

        has_sibling_pages = True
        if is_dp:
            container = aq_parent(self.obj)
            res = catalog(
                path={'query': '/'.join(container.getPhysicalPath())})
            if len([b for b in res if b.UID != self.image_to_lead]) > 1:
                has_sibling_pages = False

        try:
            state = workflow.getInfoFor(ob=self.obj, name='review_state')
        except Exception:
            state = None

        try:
            review_history = workflow.getInfoFor(
                ob=self.obj, name='review_history')
        except Exception:
            review_history = []
        for h in review_history:
            h['time'] = h['time'].HTML4()

        return {
            'portal_type': self.obj.portal_type,
            'layout': self.obj.getLayout(),
            'uid': get_uid(self.obj),
            'is_default_page': is_dp,
            'state': state,
            'review_history': review_history,
            'has_sibling_pages': has_sibling_pages
        }


class ArchetypesExporter(ContentExporter):

    def get_field_data(self):
        data = {}
        if not hasattr(self.obj, 'Schema'):
            print('No schema on {}'.format(self.obj))
            return {}
        for field in self.obj.Schema().fields():
            try:
                fdata = field.getRaw(self.obj)
            except ValueError:
                continue
            if type(fdata) == ImplicitAcquisitionWrapper:
                fdata = fdata.aq_base
            data[field.__name__] = fdata
            if field.__name__ in ('image', 'file'):
                data[field.__name__ + '_filename'] = field.getFilename(
                    self.obj)
                data[field.__name__ + '_contentType'] = field.getContentType(
                    self.obj)
            if field.__name__ == 'image':
                im = field.get(self.obj)
                if im:
                    self.im_width = im.width
        return data

    def get_referenced_images(self, data):
        images = super(ArchetypesExporter, self).get_referenced_images(data)
        if self.im_width < 200:
            if len(images) > 0 and images[0].portal_type == 'Image':
                image = images[0]
                if len(image.getBackReferences(relationship='isReferencing')) < 2:  # noqa
                    data.update({
                        'image': image.Schema().getField('image').getRaw(image),  # noqa
                        'image_filename': image.getFilename(),
                        'image_contentType': image.getContentType(),
                        'imageCaption': image.Description()
                    })
                    self.image_to_lead = get_uid(image)
                    images = images[1:]
        return images


class DexterityExporter(ContentExporter):

    def get_field_data(self):
        from plone.dexterity.interfaces import IDexterityFTI
        from plone.behavior.interfaces import IBehaviorAssignable

        data = {}

        portal_type = self.obj.portal_type
        schema = getUtility(IDexterityFTI, name=portal_type).lookupSchema()
        for name, field in getFieldsInOrder(schema):
            if self.obj.portal_type == 'File' and name == 'file':
                fileobj = getattr(self.obj, name, None)
                if fileobj is not None:
                    data[name] = base64.b64encode(fileobj.data)
            else:
                data[name] = getattr(self.obj, name, None)

        behavior_assignable = IBehaviorAssignable(self.obj)
        for behavior in behavior_assignable.enumerateBehaviors():
            binst = behavior.interface(self.obj)
            bdata = {}
            for name, field in getFieldsInOrder(behavior.interface):
                if isinstance(field, RichText):
                    textfield = getattr(binst, name, None)
                    if textfield is not None:
                        bdata[name] = textfield.raw
                else:
                    bdata[name] = getattr(binst, name, None)
            data[behavior.interface.__identifier__] = bdata

        if ILayoutAware.providedBy(self.obj):
            from plone.tiles.data import ANNOTATIONS_KEY_PREFIX
            from plone.app.blocks.utils import getLayout
            from repoze.xmliter.utils import getHTMLSerializer
            from plone.app.blocks import tiles
            from plone.app.blocks import gridsystem
            from lxml.html import tostring
            tdata = {}
            annotations = IAnnotations(self.obj, {})
            for key in annotations.keys():
                if key.startswith(ANNOTATIONS_KEY_PREFIX):
                    adata = annotations[key]
                    tdata[key] = adata
            data['tile_data'] = tdata

            req = site.REQUEST
            layout = getLayout(self.obj)
            dom = getHTMLSerializer(layout)

            try:
                tiles.renderTiles(
                    req, dom.tree, site=site,
                    baseURL=self.obj.absolute_url() + '/layout_view')
            except TypeError:
                try:
                    tiles.renderTiles(
                        req, dom.tree,
                        baseURL=self.obj.absolute_url() + '/layout_view')
                except Exception:
                    tiles.renderTiles(req, dom.tree)
            gridsystem.merge(req, dom.tree)

            data['rendered_layout'] = tostring(dom.tree)

        return data


def export_archetype_obj(obj):

    exporter = ArchetypesExporter(obj)
    data = exporter.get_field_data()
    images = exporter.get_referenced_images(data)

    if exporter.image_to_lead and 'text' in data:
        data['text'] = data['text'].replace(
            exporter.image_to_lead, get_uid(obj))

    alldata = exporter.get_object_data()
    alldata['data'] = data

    yield obj, alldata
    for ref in images:
        if get_uid(ref) == get_uid(obj):
            continue
        if ref.portal_type != 'Image':
            continue
        if exporter.image_to_lead == get_uid(ref):
            continue
        for o, d in export_obj(ref):
            if get_uid(o) == get_uid(ref) or get_uid(o) == get_uid(obj):
                continue
            yield o, d


def export_dexterity_obj(obj):
    exporter = DexterityExporter(obj)
    data = exporter.get_field_data()
    alldata = exporter.get_object_data()
    alldata['data'] = data
    yield obj, alldata


def export_obj(obj):
    if IDexterityContent.providedBy(obj):
        print("--> Dexterity: %s" % obj.Title())
        func = export_dexterity_obj
    else:
        print("--> Archetypes: %s" % obj.Title())
        func = export_archetype_obj
    for result in func(obj):
        yield result


def write_export(obj, data):
    objpath = '/'.join(obj.getPhysicalPath())
    if IFolder.providedBy(obj):
        objpath = os.path.join(objpath, '__folder__')
    else:
        fobj = aq_parent(obj)
        while not ISiteRoot.providedBy(fobj):
            if not os.path.exists(
                    os.path.join('/'.join(fobj.getPhysicalPath()), '__folder__')):  # noqa
                for eobj, edata in export_obj(fobj):
                    write_export(eobj, edata)
            fobj = aq_parent(fobj)
    path = create_path(objpath)
    try:
        fi = open(path, 'w')
        fi.write(dumps(data))
        fi.close()
    except UnicodeDecodeError:
        print('Error exporting {}'.format(objpath))


def run_export(brains):
    size = len(brains)
    for idx, brain in enumerate(brains):
        try:
            interval = int(args.interval)
            time.sleep(interval)
        except Exception:
            pass
        path = brain.getPath()
        if (args.path_filter and
                not fnmatch(path, args.path_filter)):
            print('skipping(filtered), ', path,
                  ' ', str(idx + 1) + '/' + str(size))
            continue
        print('processing, ', path, ' ', str(idx + 1) + '/' + str(size))
        try:
            obj = brain.getObject()
        except Exception:
            print('skipping - error getting object, ', path, ' ',
                  str(idx + 1) + '/' + str(size))
            continue
        for obj, data in export_obj(obj):
            write_export(obj, data)


if args.createdsince:
    print('exporting items created since %s' % args.createdsince)
    date_range = {
        'query': (
            DateTime(args.createdsince),
            DateTime('2062-05-08 23:59:59'),
        ),
        'range': 'min:max'
    }
    query = catalog(created=date_range)
elif args.modifiedsince:
    print('exporting items modified since %s' % args.modifiedsince)
    date_range = {
        'query': (
            DateTime(args.modifiedsince),
            DateTime('2062-05-08 23:59:59'),
        ),
        'range': 'min:max'
    }
    query = catalog(modified=date_range)
else:
    query = catalog()
run_export(query)



import base64
import logging
from cStringIO import StringIO
from os import fstat

import zope.publisher.interfaces
from Acquisition import aq_inner
from castle.cms.files import aws
from castle.cms.interfaces import IReferenceNamedImage
from PIL import Image
from plone import api
from plone.app.blob.download import handleRequestRange
from plone.app.blob.iterators import BlobStreamIterator
from plone.app.blob.utils import openBlob
from plone.app.contenttypes.browser import file
from plone.app.imaging.utils import getAllowedSizes
from plone.namedfile import browser as namedfile
from plone.namedfile.interfaces import INamedBlobFile
from plone.namedfile.scaling import ImageScaling
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.MimetypesRegistry.MimeTypeItem import guess_icon_path
from webdav.common import rfc1123_date
from zExceptions import NotFound
from ZODB.POSException import POSKeyError
from zope.component import getMultiAdapter


logger = logging.getLogger('castle.cms')


class DownloadAsPNG(BrowserView):
    base64 = False

    def get_data(self):
        if not self.context.image:
            raise NotFound

        blob = self.context.image._blob
        return openBlob(blob)

    def __call__(self):
        fi = self.get_data()
        im = Image.open(fi)

        changed = StringIO()
        im.save(changed, 'PNG')
        changed.seek(0)
        if hasattr(fi, 'close'):
            fi.close()
        data = changed.read()

        if self.base64:
            data = base64.b64encode(data)

        resp = self.request.response
        resp.setHeader('Content-Disposition',
                       'inline; filename=%s.png' % self.context.getId())
        resp.setHeader("Content-Length", data)
        resp.setHeader('Content-Type', 'image/png')
        return data


class ConvertToPNG(DownloadAsPNG):
    base64 = True

    def get_data(self):
        return self.request.form['data']


class Download(namedfile.Download):

    def serve_file(self):
        file = self._getFile()
        self.set_headers(file)
        if not INamedBlobFile.providedBy(file):
            return super(Download, self).__call__()

        request_range = handleRequestRange(
            self.context, file.getSize(), self.request, self.request.response)
        return BlobStreamIterator(file._blob, **request_range)

    def __call__(self):
        if not aws.uploaded(self.context):
            try:
                return self.serve_file()
            except (POSKeyError, SystemError):
                # if there is no blob file, just give out 404
                raise NotFound

        url = aws.get_url(self.context)
        self.request.response.redirect(url)


class DownloadBlob(BrowserView):

    content_type = 'application/pdf'
    file_ext = 'pdf'

    def get_data(self):
        raise NotImplementedError()

    def __call__(self):
        try:
            data = self.get_data()
        except IOError:
            # can be from zeo client blob file weirdness with PIL
            # ocassionally
            logger.info('Could not get blob data', exc_info=True)
            raise NotFound

        if data:
            is_blob = False
            if isinstance(data, basestring):
                length = len(data)
            else:
                is_blob = True
                blobfi = openBlob(data)
                length = fstat(blobfi.fileno()).st_size
                blobfi.close()

            self.request.response.setHeader(
                'Last-Modified',
                rfc1123_date(self.context._p_mtime))
            resp = self.request.response
            resp.setHeader('Content-Disposition',
                           'inline; filename=%s.%s' % (self.context.getId(), self.file_ext))
            resp.setHeader("Content-Length", length)
            resp.setHeader('Content-Type', self.content_type)

            if is_blob:
                resp.setHeader('Accept-Ranges', 'bytes')
                range = handleRequestRange(
                    self.context, length, self.request,
                    self.request.response)
                return BlobStreamIterator(data, **range)
            else:
                return data
        else:
            raise NotFound


class EmptyRedirect(BrowserView):
    '''
    I just want to redirect.
    Yes, this doc string is required because Zope2 is insane
    '''

    __roles__ = ('Anonymous',)

    def __init__(self, *args):
        super(EmptyRedirect, self).__init__(*args)
        self.sizes = getAllowedSizes()

    def __call__(self):
        return self

    def __getattr__(self, name):
        if name in self.sizes:
            return self
        raise AttributeError


class CastleImageScaling(ImageScaling):

    def publishTraverse(self, request, name):
        if name == 'image':
            if IReferenceNamedImage.providedBy(self.context.image):
                # auth not setup yet, we just redirect and assume we'll figure the rest out
                catalog = api.portal.get_tool('portal_catalog')
                brains = catalog.unrestrictedSearchResults(UID=self.context.image.reference)
                if len(brains) > 0:
                    brain = brains[0]
                    url = request.ACTUAL_URL.replace(
                        self.context.absolute_url(), brain.getURL())
                    request.response.redirect(url)
                    view = EmptyRedirect(self.context, self.request)
                    return view.__of__(self.context)
        return super(CastleImageScaling, self).publishTraverse(request, name)


class NamedFileDownload(Download):
    def __call__(self):
        try:
            return super(NamedFileDownload, self).__call__()
        except AttributeError:
            raise zope.publisher.interfaces.NotFound(self, self.fieldname, self.request)


class FileView(file.FileView):

    @property
    def moved(self):
        return getattr(self.context.file, 'original_filename', None) is not None

    @property
    def filename(self):
        fi = self.context.file
        return getattr(fi, 'original_filename', fi.filename)

    @property
    def size(self):
        fi = self.context.file
        return getattr(fi, 'original_size', fi.getSize())

    @property
    def content_type(self):
        fi = self.context.file
        return getattr(fi, 'original_content_type', fi.contentType)

    def get_mimetype_icon(self):
        context = aq_inner(self.context)
        pstate = getMultiAdapter(
            (context, self.request),
            name=u'plone_portal_state'
        )
        portal_url = pstate.portal_url()
        mtr = getToolByName(context, "mimetypes_registry")
        mime = []
        if self.content_type:
            mime.append(mtr.lookup(self.content_type))
        if self.filename:
            mime.append(mtr.lookupExtension(self.filename))
        mime.append(mtr.lookup("application/octet-stream")[0])
        icon_paths = [m.icon_path for m in mime if hasattr(m, 'icon_path')]
        if icon_paths:
            return portal_url + "/" + icon_paths[0]

        return portal_url + "/" + guess_icon_path(mime[0])



from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.resources.browser.cook import cookWhenChangingSettings

PROFILE_ID = 'profile-castle.cms:2_6_2'


def upgrade(site, logger=None):
    setup = getToolByName(site, 'portal_setup')
    setup.runAllImportStepsFromProfile(PROFILE_ID)
    cookWhenChangingSettings(site)



import csv
import json
import os
from datetime import datetime
from io import BytesIO

import sqlalchemy.exc
from castle.cms import linkreporter
from castle.cms.linkreporter import Link
from castle.cms.linkreporter import Url
from Products.Five import BrowserView


class LinksControlPanel(BrowserView):

    label = 'Link report'
    description = 'Report on crawled site links'

    page_size = 20
    _broken_count = None
    _page = None

    def __call__(self):
        self.session = linkreporter.get_session()
        if 'links_of' in self.request.form:
            links = []
            for link in self.session.query(Link).filter(
                    Link.site_id == self.context.getId(),
                    Link.url_to == self.request.form['links_of']).limit(50):
                links.append(link.url_from)
            self.request.response.setHeader('Content-type', 'application/json')
            resp = json.dumps(links)
        elif self.request.form.get('export') == 'true':
            resp = self.csv_export()
        else:
            resp = super(LinksControlPanel, self).__call__()
        self.session.close()
        return resp

    @property
    def configured(self):
        return os.environ.get('LINK_REPORT_DB', 'sqlite://') != 'sqlite://'

    def csv_export(self):
        output = BytesIO()
        writer = csv.writer(output)

        writer.writerow(['URL', 'From URL', 'Status code', 'Checked', 'Redirected url'])
        last = None
        count = 0
        for link, url in self.session.query(Link, Url.url).join(
                Url, Url.url == Link.url_to).filter(
                    Link.site_id == self.context.getId(),
                    Url.last_checked_date > datetime(1985, 1, 1),
                    ~Url.status_code.in_([200, 999, 429, 524]),
                ).order_by(Url.url):

            if last != link.url_to:
                if count > 25:
                    writer.writerow(
                        [last, '- Truncated {} entries...'.format(count - 25)])
                last = link.url_to
                count = 0
            count += 1
            if count <= 25:
                writer.writerow([
                    link.url_to,
                    link.url_from,
                    linkreporter.status_codes_info.get(
                        link._url_to.status_code, link._url_to.status_code),
                    link._url_to.last_checked_date.isoformat(),
                    link._url_from.final_url or ''
                ])

        writer.writerow([])
        writer.writerow([])
        writer.writerow([])

        for name, value in self.get_summary().items():
            if name not in ('errors',):
                writer.writerow([name, value, ''])

        resp = self.request.response
        resp.setHeader('Content-Disposition', 'attachment; filename=broken-links.csv')
        resp.setHeader('Content-Type', 'text/csv')
        output.seek(0)
        return output.read()

    def get_broken_query(self):
        return self.session.query(Link, Url.url).join(
            Url, Link.url_to == Url.url).filter(
                Link.site_id == self.context.getId(),
                ~Url.status_code.in_([-1, 200, 999, 429, 524])
        )

    @property
    def broken_count(self):
        if self._broken_count is not None:
            return self._broken_count

        query = self.get_broken_query()
        try:
            self._broken_count = query.count()
        except sqlalchemy.exc.OperationalError as ex:
            if 'no such table' in ex.message:
                linkreporter.init()
                self._broken_count = query.count()
            raise
        return self._broken_count

    @property
    def page(self):
        try:
            return int(self.request.form.get('page', '0'))
        except Exception:
            return 0

    def get_summary(self):
        try:
            return {
                'links': '{:,}'.format(self.session.query(Link).filter(
                    Link.site_id == self.context.getId()
                ).count()),
                'urls': '{:,}'.format(self.session.query(Url, Link.url_to).join(
                    Link, Link.url_to == Url.url).filter(
                        Link.site_id == self.context.getId()
                    ).distinct(Url.url).count()),
                'unchecked': '{:,}'.format(self.session.query(Url, Link.url_to).join(
                    Link, Link.url_to == Url.url).filter(
                        Link.site_id == self.context.getId(),
                        Url.status_code == -1,
                        Url.parse_error == None  # noqa
                    ).count()),
                'errors': '{:,}'.format(self.session.query(Url, Link.url_to).join(
                    Link, Link.url_to == Url.url).filter(
                        Link.site_id == self.context.getId(),
                        Url.status_code.in_([-2, -200, -599, -301]),
                        Url.parse_error != None
                    ).count()),
                'broken': '{:,}'.format(self.broken_count),
            }
        except sqlalchemy.exc.OperationalError as ex:
            if 'no such table' in ex.message:
                linkreporter.init()
                return self.get_summary()
            raise

    def get_broken(self):
        offset = self.page * self.page_size
        items = [l[0] for l in self.get_broken_query().order_by(
                 Url.last_checked_date.desc()).offset(offset).limit(self.page_size)]
        return {
            'items': items,
            'total': self.broken_count
        }



# -*- coding: utf-8 -*-
from castle.cms._scripts.cleanup_archive import is_s3_url
from castle.cms import archival
from castle.cms.testing import CASTLE_PLONE_FUNCTIONAL_TESTING
from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from moto import mock_s3

import unittest
from urlparse import urlparse


class TestCleanupArchive(unittest.TestCase):
    layer = CASTLE_PLONE_FUNCTIONAL_TESTING

    def setUp(self):
        self.test_access_key = u'AKIAIOSFODNN7EXAMPLE'
        self.test_secret_key = u'wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY'
        self.test_bucket_name = u'castletest'
        # possibly useful to set under some condition...
        self.test_bucket_endpoint = u'https://s3.amazonaws.com/castletest'
        self.test_base_url = u'https://localhost.localdomain/'  # used for test swap
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
        api.portal.set_registry_record('castle.aws_s3_key', self.test_access_key)
        api.portal.set_registry_record('castle.aws_s3_secret', self.test_secret_key)
        api.portal.set_registry_record('castle.aws_s3_bucket_name', self.test_bucket_name)
        api.portal.set_registry_record('castle.aws_s3_host_endpoint', self.test_bucket_endpoint)
        api.portal.set_registry_record('castle.aws_s3_base_url', self.test_base_url)

    def tearDown(self):
        pass

    @mock_s3
    def test_is_s3_url(self):
        site = api.portal.get()
        storage = archival.Storage(site)
        parsed_endpoint = urlparse(storage.s3_conn.meta.client.meta.endpoint_url)

        urlstr = "http://localhost:9000/{}/{}".format(self.test_bucket_name, self.test_access_key)
        self.assertFalse(is_s3_url(urlstr, parsed_endpoint))

        urlstr = "https://s3.amazonaws.com/{}/{}".format(self.test_bucket_name, self.test_access_key)
        self.assertTrue(is_s3_url(urlstr, parsed_endpoint))



from castle.cms.widgets import SelectFieldWidget
from plone.autoform import directives
from plone.autoform.interfaces import IFormFieldProvider
from plone.supermodel import model
from zope import schema
from zope.interface import alsoProvides
from zope.schema.vocabulary import SimpleVocabulary


class ISearch(model.Schema):

    model.fieldset(
        'settings',
        fields=['searchterm_pins', 'robot_configuration'],
    )

    searchterm_pins = schema.List(
        title=u'Search Term Pins',
        description=u'Searches this content should show top in results for.',
        required=False,
        value_type=schema.TextLine(),
        default=[]
    )

    directives.widget('robot_configuration', SelectFieldWidget)
    robot_configuration = schema.List(
        title=u'How robots should behave with this content',
        description=u'Robot index and consume content on the web. '
                    u'This allows you to remove content from search indexes.',
        required=False,
        defaultFactory=lambda: ['index', 'follow'],
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary([
                SimpleVocabulary.createTerm('index', 'index', 'Index'),
                SimpleVocabulary.createTerm('follow', 'follow', 'Follow links'),
                SimpleVocabulary.createTerm('noimageindex', 'noimageindex', 'Do not index images'),
                SimpleVocabulary.createTerm(
                    'noarchive', 'noarchive',
                    'Search engines should not show a cached link '
                    'to this page on a SERP'),
                SimpleVocabulary.createTerm(
                    'nosnippet', 'nosnippet',
                    'Tells a search engine not to show a snippet of this page '
                    '(i.e. meta description) of this page on a SERP.'),
            ])
        )
    )


alsoProvides(ISearch, IFormFieldProvider)



import os
import subprocess

import celery.bin.celery
import elasticsearch
import redis
from collective.celery.utils import getCelery
from collective.elasticsearch.es import ElasticSearchCatalog
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView


class StatusView(BrowserView):

    def docsplit(self):
        try:
            subprocess.check_call('docsplit')
            return True, 'ok'
        except OSError as e:
            return False, str(e)

    def redis(self):
        server = os.environ.get("REDIS_SERVER", "127.0.0.1:6379")
        host, _, port = server.partition(':')
        if not host or not port:
            return '[{}] {}'.format(server, 'Invalid configuration')

        client = redis.StrictRedis(
            host=host, port=int(port), db=0,
            socket_timeout=0.5, socket_connect_timeout=0.5)
        try:
            client.client_list()
            return True, '[{}] ok'.format(server)
        except redis.ConnectionError as e:
            return False, '[{}] {}'.format(server, str(e))

    def celery(self):
        app = getCelery()
        status = celery.bin.celery.CeleryCommand.commands['status']()
        status.app = app
        try:
            status.run()
            return True, 'ok'
        except Exception as e:
            return False, str(e)

    def elasticsearch(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        es = ElasticSearchCatalog(catalog)
        conn = es.connection
        try:
            (conn.cluster.health())
            return True, 'ok'
        except elasticsearch.ConnectionError as e:
            return False, str(e)



from zope.publisher.interfaces import IPublishTraverse
from ZPublisher.BaseRequest import DefaultPublishTraverse
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.publisher.interfaces.http import IHTTPRequest
from zope.component import adapts
from zope.interface import implements
from plone.memoize.forever import memoize
from plone import api


class IsolatedSiteTraverser(DefaultPublishTraverse):
    """
    This traverser is applied only once you traverse an IIsolatedObject.
    This traverser look first if one of the names to be traversed matches
    with one of the root isolated objects. Then
        * if None of them match then it use the default traverser (see DefaultPublishTraverse) ;
        * if one of them match, we mark the request as being a potentially bad
          request and check if the first name we are traversing is one of the
          isolated objects.
    """
    adapts(IPloneSiteRoot, IHTTPRequest)
    implements(IPublishTraverse)

    @memoize
    def get_site_objects(self):
        """
        Return the objects on the Zope root (we are assuming here that your sites
        are on the top level) that have to be isolated from each others
        """
        return frozenset([id for id, obj in api.portal.get().aq_parent.objectItems() if IPloneSiteRoot.providedBy(obj)])  # noqa

    def publishTraverse(self, request, name):
        namesToTraverse = frozenset([name] + self.request['TraversalRequestNameStack'])
        sites = self.get_site_objects()
        if namesToTraverse.intersection(sites) and name in sites:
            raise KeyError(name)
        else:
            return super(IsolatedSiteTraverser, self).publishTraverse(request, name)



from collective.elasticsearch.es import ElasticSearchCatalog
from collective.elasticsearch.hook import index_batch
from collective.elasticsearch.interfaces import IElasticSettings
from elasticsearch import Elasticsearch
from plone import api
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from zope.component import getUtility


def ESConnectionFactoryFactory(registry=None):
    if registry is None:
        registry = getUtility(IRegistry)
    settings = registry.forInterface(IElasticSettings, check=False)
    hosts = settings.hosts
    opts = dict(
        timeout=getattr(settings, 'timeout', 0.5),
        sniff_on_start=getattr(settings, 'sniff_on_start', False),
        sniff_on_connection_fail=getattr(
            settings, 'sniff_on_connection_fail', False),
        sniffer_timeout=getattr(settings, 'sniffer_timeout', 0.1),
        retry_on_timeout=getattr(settings, 'retry_on_timeout', False)
    )

    def factory():
        return Elasticsearch(hosts, **opts)
    return factory


def index_in_es(obj):
    catalog = api.portal.get_tool('portal_catalog')
    es = ElasticSearchCatalog(catalog)
    if es.enabled:
        index_batch([], {IUUID(obj): obj}, [], es)


def add_indexes(indexes):
    """
    indexes should be a tuple of (name of the index, index type)
    """
    catalog = api.portal.get_tool('portal_catalog')
    for name, _type in indexes.items():
        if name not in catalog.indexes():
            if type(_type) == dict:
                real_type = _type['type']
                del _type['type']
                catalog.addIndex(name, real_type, **_type)
            else:
                catalog.addIndex(name, _type)


def delete_indexes(indexes):
    catalog = api.portal.get_tool('portal_catalog')
    for name in indexes:
        if name in catalog.indexes():
            catalog.delIndex(name)


def add_metadata(metadata):
    catalog = api.portal.get_tool('portal_catalog')
    _catalog = catalog._catalog
    for name in metadata:
        if name not in catalog.schema():
            # override how this works normally to not cause a reindex
            schema = _catalog.schema
            names = list(_catalog.names)
            values = schema.values()
            if values:
                schema[name] = max(values) + 1
            else:
                schema[name] = 0
            names.append(name)

            _catalog.names = tuple(names)
            _catalog.schema = schema


def delete_metadata(metadata):
    catalog = api.portal.get_tool('portal_catalog')
    for name in metadata:
        if name in catalog.schema():
            catalog.delColumn(name)



# -*- coding: utf-8 -*-
from castle.cms.tiles.base import ContentTile  # noqa b/w
from castle.cms.widgets import ImageRelatedItemsFieldWidget
from castle.cms.widgets import QueryFieldWidget
from castle.cms.widgets import RelatedItemFieldWidget
from castle.cms.widgets import UseQueryWidget
from plone.autoform import directives as form
from plone.supermodel import model
from zope import schema
from zope.interface import Invalid
from zope.interface import invariant


class IContentTileSchema(model.Schema):

    form.widget(use_query=UseQueryWidget)
    use_query = schema.Bool(
        title=u'Use dynamic query',
        default=False)

    form.widget(content=RelatedItemFieldWidget)
    content = schema.List(
        title=u"Item",
        description=u"Selected item to display",
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        )
    )

    @invariant
    def validate_content(data):
        if data.content and len(data.content) != 1:
            raise Invalid("Must select 1 item")

    form.widget(query=QueryFieldWidget)
    query = schema.List(
        title=u'Search terms',
        description=u"Define the search terms for the item you want "
                    u"to list by choosing what to match on. "
                    u"The list of results will be dynamically updated",
        value_type=schema.Dict(value_type=schema.Field(),
                               key_type=schema.TextLine()),
        required=False
    )

    sort_on = schema.TextLine(
        title=u'Sort on',
        description=u"Sort on this index",
        required=False,
    )

    sort_reversed = schema.Bool(
        title=u'Reversed order',
        description=u'Sort the results in reversed order',
        required=False,
    )

    more_text = schema.TextLine(
        title=u'More text',
        default=u'',
        required=False)


class IImagesTileSchema(model.Schema):

    form.widget(use_query=UseQueryWidget)
    use_query = schema.Bool(
        title=u'Use dynamic query',
        default=False)

    form.widget('images', ImageRelatedItemsFieldWidget)
    images = schema.List(
        title=u"Images",
        description=u"Select images or folders of images to display in "
                    u"gallery",
        value_type=schema.Choice(
            vocabulary='plone.app.vocabularies.Catalog'
        ),
        required=False
    )

    form.widget(query=QueryFieldWidget)
    query = schema.List(
        title=u'Search terms',
        description=u"Define the search terms for the images you want to use. "
                    u"The list of results will be dynamically updated",
        value_type=schema.Dict(value_type=schema.Field(),
                               key_type=schema.TextLine()),
        required=False
    )

    sort_on = schema.TextLine(
        title=u'Sort on',
        description=u"Sort on this index",
        required=False,
    )

    sort_reversed = schema.Bool(
        title=u'Reversed order',
        description=u'Sort the results in reversed order',
        required=False,
    )



# -*- coding: utf-8 -*-
import logging
import os
import sys
import time
import traceback

import Globals
from DateTime import DateTime
from plone import api
from plone.app.theming.interfaces import THEME_RESOURCE_NAME
from plone.app.theming.utils import getCurrentTheme
from plone.registry.interfaces import IRegistry
from plone.resource.utils import queryResourceDirectory
from z3c.jbot import interfaces
from z3c.jbot.manager import find_package
from zope import interface
from zope.component import ComponentLookupError
from zope.component import queryUtility
from zope.component.hooks import getSite
from zope.filerepresentation.interfaces import IRawReadFile
from zope.globalrequest import getRequest
from zope.interface import providedBy


REQ_CACHE_KEY = 'castle.cms.jbot.storage'

logger = logging.getLogger('castle.cms')
IGNORE = object()


class Storage(object):

    def __init__(self):
        self.theme = getCurrentTheme()
        self.site = getSite()
        if self.theme:
            directory = queryResourceDirectory(
                THEME_RESOURCE_NAME, self.theme)
            try:
                self.jbot_directory = directory['jbot']
            except Exception:
                self.jbot_directory = None
        else:
            self.jbot_directory = None

    def __contains__(self, name):
        if self.jbot_directory is not None:
            if name in self.jbot_directory:
                return True
        return False

    def _get_fs_path(self):
        """ returns FS path for jbot storage.
        """
        base_path = '/'.join(Globals.data_dir.split('/')[:-1])
        jbot_dir = os.path.join(base_path, 'jbot')
        site_id = '-'.join(self.site.getPhysicalPath()[1:])
        fs_path = os.path.join(jbot_dir, site_id, self.theme)
        if not os.path.exists(fs_path):
            os.makedirs(fs_path)
            logger.info('created missing directory {}'.format(fs_path))
        return fs_path

    def get_filepath(self, filename):
        modified = self.get_modified(filename)
        path = self._get_fs_path()
        return os.path.join(path, '{}-{}'.format(
            int(modified), filename))

    def get_customizations(self):
        files = {}
        if self.jbot_directory is not None:
            for filename in self.jbot_directory.listDirectory():
                filepath = self.get_filepath(filename)
                if self.file_modified(filename, filepath):
                    self.write_customization(filename)
                files[filename] = {
                    'modified': self.get_modified(filename),
                    'filepath': filepath
                }
        return files

    def write_customization(self, filename):
        fi = self.jbot_directory[filename]
        filepath = self.get_filepath(filename)
        with open(filepath, 'wb') as tmpfi:
            if fi.__class__.__name__ == "FilesystemFile":
                data = IRawReadFile(fi).read()
            else:
                data = str(fi.data)
            tmpfi.write(data)
        return filepath

    def get_modified(self, filename):
        fi = self.jbot_directory[filename]
        if fi.__class__.__name__ == "FilesystemFile":
            return fi.lastModifiedTimestamp
        else:
            return fi._p_mtime

    def file_modified(self, filename, filepath):
        return (not os.path.exists(filepath) or
                DateTime(self.get_modified(filename)) > DateTime(os.stat(filepath).st_mtime))


class _TemplateManager(object):
    interface.implements(interfaces.ITemplateManager)

    def __init__(self):
        self.syspaths = tuple(sys.path)
        self.templates = {}
        self.cache = {}

    def get_storage(self):
        try:
            return Storage()
        except ComponentLookupError:
            return None

    def _get_customizations(self):
        storage = self.get_storage()
        if storage is None:
            return {}
        else:
            return storage.get_customizations()

    def get_customizations(self):
        req = getRequest()
        key = '{}.customizations'.format(REQ_CACHE_KEY)
        if key not in req.environ:
            if api.env.debug_mode():
                req.environ[key] = self._get_customizations()
            else:
                # check for value in cache
                cache_key = providedBy(req)  # layer is cache key
                if cache_key not in self.cache:
                    self.cache[cache_key] = {
                        'when': time.time(),
                        'value': self._get_customizations()
                    }
                else:
                    registry = queryUtility(IRegistry)
                    theme_cache_time = getattr(registry, '_theme_cache_mtime', 0)
                    if theme_cache_time > self.cache[cache_key]['when']:
                        self.cache[cache_key] = {
                            'when': time.time(),
                            'value': self._get_customizations()
                        }
                req.environ[key] = self.cache[cache_key]['value']
        return req.environ[key]

    def registerTemplate(self, template, token):  # noqa
        """
        wrap this method so we can log errors easily
        """
        try:
            return self._registerTemplate(template, token)
        except Exception:
            req = getRequest()
            if 'jbot.error' not in req.environ:
                # only log one error per request
                req.environ['jbot.error'] = True
                logger.error(
                    'collective.jbot error: {}'.format(traceback.format_exc()))

    def _registerTemplate(self, template, original):  # noqa
        """
        Return True if there has been a change to the override.
        """
        # check if template is ignored, no filename to override
        filename = self.templates.get(template)
        if filename is IGNORE:
            return

        customized = False
        if template in self.templates:
            # already customized by us
            customized = True
            filename = self.templates[template]['filename']
        else:
            # check if an overridable resource
            path = find_package(self.syspaths, template.filename)
            if path is None:
                # permanently ignore template
                self.templates[template] = IGNORE
                return
            filename = path.replace(os.path.sep, '.')

        customizations = self.get_customizations()
        if filename in customizations:
            filepath = customizations[filename]['filepath']
            if template not in self.templates:
                # not cached, save it
                self.templates[template] = {
                    'filename': filename,
                    'filepath': filepath,
                    'original': template.filename
                }
            else:
                self.templates[template]['filepath'] = filepath

            if template.filename != filepath:
                # if filename not same as customization, edit and return True
                template.filename = filepath
                return True
        else:
            if customized and template in self.templates:
                # revert, we had it customized but it's no longer customizations
                template.filename = self.templates[template]['original']
                del self.templates[template]
                return True


class TemplateManagerFactory(object):
    interface.implements(interfaces.ITemplateManager)

    def __init__(self):
        self.manager = _TemplateManager()

    def __call__(self, layer):
        return self.manager


TemplateManager = TemplateManagerFactory()


def mark_request(site, event):
    '''
    Mark request so each site has a unique layer `Provides` object.
    This allows us to use jbot with themes correctly without
    needing to fork/patch z3c.jbot(which would be nasty)
    '''
    if getattr(site, '_v_unique_jbot_layer', None) is None:

        class IUniqueSiteLayer(interface.Interface):
            pass

        site._v_unique_jbot_layer = IUniqueSiteLayer

    interface.alsoProvides(event.request, site._v_unique_jbot_layer)



import hashlib
import json
import random
import time
from datetime import datetime
from hashlib import sha256 as sha

from BTrees.OOBTree import OOBTree
from DateTime import DateTime
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from plone import api
from ZODB.POSException import ConflictError


try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    using_sysrandom = False


SECRET = random.randint(0, 1000000)


def get_random_string(length=12,
                      allowed_chars='abcdefghijklmnopqrstuvwxyz'
                                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(
            sha(
                "%s%s%s" % (
                    random.getstate(),
                    time.time(),
                    SECRET)
                ).digest())  # noqa
    return ''.join([random.choice(allowed_chars) for i in range(length)])


def make_random_key(length=150, prefix=''):
    if prefix:
        prefix = str(prefix) + '-'
    hashed = '%s%s' % (
        hashlib.sha1(str(get_random_string(length)).encode('utf-8')).hexdigest()[:5],  # noqa
        str(datetime.now().microsecond)
    )
    return prefix + hashlib.sha1(hashed.encode('utf-8')).hexdigest()[:length]


def retriable(count=3, sync=False, reraise=True, on_retry_exhausted=None):

    def decorator(func):
        def wrapped(*args, **kwargs):
            retried = 0
            if sync:
                try:
                    api.portal.get()._p_jar.sync()
                except Exception:
                    pass
            while retried < count:
                try:
                    return func(*args, **kwargs)
                except ConflictError:
                    retried += 1
                    try:
                        api.portal.get()._p_jar.sync()
                    except Exception:
                        if retried >= count:
                            if on_retry_exhausted is not None:
                                on_retry_exhausted(*args, **kwargs)
                            if reraise:
                                raise
        return wrapped
    return decorator


def get_ip(req):
    ip = req.get('HTTP_CF_CONNECTING_IP')
    if not ip:
        ip = req.get('HTTP_X_FORWARDED_FOR')
    if not ip:
        ip = req.get('HTTP_X_REAL_IP')
    if not ip:
        ip = req.get('REMOTE_ADDR')
    return ip


def normalize_url(url):
    url = url.split('#')[0]
    if url.startswith('data:'):
        return None
    return url


def strings_differ(string1, string2):
    """Check whether two strings differ while avoiding timing attacks.

    This function returns True if the given strings differ and False
    if they are equal.  It's careful not to leak information about *where*
    they differ as a result of its running time, which can be very important
    to avoid certain timing-related crypto attacks:

        http://seb.dbzteam.org/crypto/python-oauth-timing-hmac.pdf

    """
    if len(string1) != len(string2):
        return True

    invalid_bits = 0
    for a, b in zip(string1, string2):
        invalid_bits += a != b

    return invalid_bits != 0


def _customhandler(obj):
    if type(obj) == DateTime:
        return obj.ISO8601()
    if type(obj) == PersistentList:
        return list(obj)
    if type(obj) == PersistentDict:
        return dict(obj)
    if type(obj) == OOBTree:
        return dict(obj)
    raise TypeError(
        "Unserializable object {} of type {}".format(obj, type(obj)))


def json_dumps(data):
    return json.dumps(data, default=_customhandler)



import hmac
import json

from castle.cms.utils import get_chat_info
from plone import api
from plone.keyring.interfaces import IKeyManager
from plone.protect.authenticator import _getKeyring
from plone.protect.authenticator import _is_equal
from Products.Five import BrowserView
from zope.component import getUtility


try:
    from hashlib import sha1 as sha
except ImportError:
    import sha


class ChatLogin(BrowserView):

    def __call__(self):

        self.request.response.setHeader('Content-Type', 'application/json')

        token = self.request.get('token')
        user = self.request.get('user')

        if token:
            salt = api.portal.get_registry_record('castle.rocket_chat_secret')

            manager = getUtility(IKeyManager)
            keyring = _getKeyring(user, manager=manager)
            for key in keyring:
                if key is None:
                    continue
                value = hmac.new(key, user + salt, sha).hexdigest()
                if _is_equal(value, token):
                    return json.dumps({
                        'status': 'success',
                        'user': user
                    })
        return json.dumps({'status': 'failure'})


class ChatView(BrowserView):

    def chatInfo(self):
        return json.dumps(get_chat_info())



from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ITinyMCESchema
from Products.CMFPlone.interfaces.atd import IATDProxyView
from zope.component import getUtility
from zope.interface import implements

import requests


class ATDProxyView(object):
    """ Proxy for the 'After the Deadline' spellchecker
    """
    implements(IATDProxyView)

    def checkDocument(self):
        """ Proxy for the AtD service's checkDocument function
            See http://www.afterthedeadline.com/api.slp for more info.
        """
        registry = getUtility(IRegistry)
        settings = registry.forInterface(
            ITinyMCESchema, prefix="plone", check=False)
        if settings.libraries_spellchecker_choice != 'AtD':
            return 'atd not enabled'

        tool = getToolByName(self.context, "portal_membership")
        if bool(tool.isAnonymousUser()):
            return 'must be authenticated to use atd'

        if 'data' in self.request.form:
            data = self.request.form
        else:
            data = self.request._file.read()

        service_url = settings.libraries_atd_service_url
        if not service_url.startswith('http'):
            service_url = 'https://' + service_url

        resp = requests.post(service_url + '/checkDocument', data=data)

        if resp.status_code != 200:
            raise Exception('Unexpected response code from AtD service %d' %
                            resp.status_code)

        self.request.RESPONSE.setHeader('content-type',
                                        'text/xml;charset=utf-8')
        respxml = resp.content
        xml = respxml.strip().replace("\r", '').replace("\n", '').replace(
            '>  ', '>')
        return xml



# -*- coding: utf-8 -*-
import pytz
from castle.cms.pwexpiry.config import _
from plone import api
from Products.CMFPlone.utils import safe_unicode
from zope.component import getMultiAdapter
from zope.i18n import translate


def send_notification_email(user, days_to_expire,
                            email_view='notification_email'):
    """
    """

    language = api.portal.get_default_language()
    if user.getProperty('language'):
        language = user.getProperty('language')

    recipient = user.getProperty('email')
    portal = api.portal.get()

    email_template = getMultiAdapter(
        (portal, portal.REQUEST), name=email_view
    )

    body = email_template(**{
        'username': user.getUserId(),
        'fullname': safe_unicode(user.getProperty('fullname')),
        'days': days_to_expire,
        'language': language,
    })

    subject = translate(
        _('email_subject',
          default=u"${days} days left to password expiration",
          mapping={'days': days_to_expire},
          ),
        target_language=language
    )

    api.portal.send_email(recipient=recipient,
                          subject=subject,
                          body=body)


def days_since_event(event_date, current_date):
    """
    Returns the number of days difference
    between two given dates
    """
    # make both dates timezone aware
    if not event_date.tzinfo:
        event_date = pytz.utc.localize(event_date)
    if not current_date.tzinfo:
        current_date = pytz.utc.localize(current_date)

    difference = current_date - event_date
    return difference.days



# -*- coding: utf-8 -*-

from zope.component.interfaces import IObjectEvent
from zope.interface import Attribute, Interface


class ICustomPasswordValidator(Interface):
    """
    Defines custom validators for the password
    chosen by the users.
    """

    def validate(password):
        """
        Validates password
        """


class IValidPasswordEntered(IObjectEvent):
    """
    """


class IInvalidPasswordEntered(IObjectEvent):
    """
    """


class IUserUnlocked(Interface):
    """
    """


user = Attribute("The user that was unlocked")



from castle.cms.interfaces import ISecureLoginAllowedView
from castle.cms.interfaces import IAuthenticator
from Products.PasswordResetTool.PasswordResetTool import ExpiredRequestError
from Products.PasswordResetTool.PasswordResetTool import InvalidRequestError
from plone.protect.interfaces import IDisableCSRFProtection
from Products.Five import BrowserView
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implements
from zope.i18n import translate
from plone import api

import json


class PasswordResetView(BrowserView):
    implements(ISecureLoginAllowedView)

    def __init__(self, context, request):
        super(PasswordResetView, self).__init__(context, request)
        self.auth = self.authenticator = getMultiAdapter(
            (context, request), IAuthenticator)

    def __call__(self):
        if self.request.REQUEST_METHOD == 'POST':
            self.request.response.setHeader('Content-type', 'application/json')
            return self.reset_password()
        else:
            return self.index()

    def reset_password(self):
        pw_tool = api.portal.get_tool('portal_password_reset')
        registration = api.portal.get_tool('portal_registration')
        userid = self.request.form.get('userid')
        randomstring = self.request.form.get('code')
        password = self.request.form.get('password')

        err_str = registration.testPasswordValidity(password)
        if err_str is not None:
            return json.dumps({
                'success': False,
                'message': translate(err_str)
            })

        alsoProvides(self.request, IDisableCSRFProtection)
        try:
            pw_tool.resetPassword(userid, randomstring, password)
        except ExpiredRequestError:
            return json.dumps({
                'success': False,
                'message': 'Password reset request has expired'
            })
        except (InvalidRequestError, RuntimeError):
            return json.dumps({
                'success': False,
                'message': 'Password reset request is invalid'
            })

        return json.dumps({
            'success': True,
            'message': 'Password reset successfully'
        })

    def get_data(self):
        data = {
            'code': self.request.form.get('code'),
            'userid': self.request.form.get('userid'),
            'apiEndpoint': self.context.absolute_url() + '/@@password-reset',
            'successUrl': api.portal.get().absolute_url() + '/@@secure-login'
        }
        return json.dumps(data)



from castle.cms import subscribe
from plone.stringinterp import _
from plone.stringinterp.adapters import BaseSubstitution
from plone.stringinterp.dollarReplace import LazyDict
from plone.stringinterp.interfaces import IStringInterpolator
from Products.CMFCore.interfaces import IContentish
import string
from zope.component import adapts, adapter
from zope.interface import implementer


@adapter(IContentish)
class SubscriberMailAddressSubstitution(BaseSubstitution):
    pass


class SubscriberCategoryEmailSubstitution(BaseSubstitution):
    category = _(u'Category Subscribers')
    description = _(
        u'Subscriber E-Mail addresses for category "x", managed under '
        u'Site Setup -> Announcements -> Manage Categories (ex: '
        u'"category_emails_for_news"). NOTE: "x" is case-insensitive and '
        u'spaces should be replaced with underscores, ex: "Meeting Minutes" '
        u'would be "${category_emails_for_meeting_minutes}"')

    def __init__(self, context, category=None):
        super(SubscriberCategoryEmailSubstitution, self).__init__(context)
        self.category = category.lower().replace(" ", "_")

    def safe_call(self):
        emails = []
        for subscriber in subscribe.all():
            if 'categories' not in subscriber:
                continue
            categories = subscriber['categories']
            for c in categories:
                if c.lower().replace(" ", "_") == self.category:
                    email = subscriber.get("email", None)
                    if email is not None:
                        emails.append(email)
                    break
        return u', '.join(emails)


class CastleLazyDict(LazyDict):
    def __getitem__(self, key):
        if key.startswith("category_emails_for_"):
            category = key[20:]
            res = self._cache.get(key)
            if res is None:
                res = SubscriberCategoryEmailSubstitution(self.context, category=category)()
            return res
        else:
            return super(CastleLazyDict, self).__getitem__(key)


@implementer(IStringInterpolator)
class CastleInterpolator(object):
    adapts(IContentish)

    def __init__(self, context):
        self._cldict = CastleLazyDict(context)

    def __call__(self, s):
        return string.Template(s).safe_substitute(self._cldict)



from zope.component.hooks import setSite
import argparse


parser = argparse.ArgumentParser(
    description='...')
parser.add_argument('--site-id', dest='site_id', default='Plone')
args, _ = parser.parse_known_args()


if __name__ == '__main__':
    site = app[args.site_id]  # noqa
    setSite(site)

    catalog = site.portal_catalog
    brains = catalog()

    latest_modified = brains[0].modified

    for idx, brain in enumerate(brains):
        try:
            latest_modified = brain.modified if brain.modified > latest_modified else latest_modified  # noqa
        except Exception:
            print('No modified date on brain {path}'.format(path=brain.getPath()))

    print('Content last modified {latest}'.format(latest=latest_modified))



# overridable default settings from environment variables
import os


def get(name, default, _type='unicode'):
    env_name = 'DEFAULT_{}'.format(name.upper())
    if env_name in os.environ:
        value = os.environ[env_name]
    else:
        value = default
    if _type == 'unicode':
        if not isinstance(value, unicode):
            value = unicode(value)
    elif _type == 'bool':
        if not isinstance(value, bool):
            value = value.lower() in ('t', 'true', '1', 'on')
    return value



import json
import logging
import mimetypes
import time

from apiclient.http import MediaFileUpload
from castle.cms.interfaces import IUploadedToYoutube
from collective.celery.utils import getCelery
from plone.registry.interfaces import IRegistry
from zope.component import getUtility
from zope.interface import alsoProvides

from . import get_service


logger = logging.getLogger('castle.cms')

DEFAULT_CHUNK_SIZE = -1  # one request, still resumable
MAX_RETRIES = 3
PRIVATE_VIDEO_STATUS = 'unlisted'

SCOPES = [
    'https://www.googleapis.com/auth/youtube',
    'https://www.googleapis.com/auth/youtube.force-ssl',
    'https://www.googleapis.com/auth/youtube.upload',
    'https://www.googleapis.com/auth/youtube.readonly',
    'https://www.googleapis.com/auth/yt-analytics.readonly',
    'https://www.googleapis.com/auth/youtubepartner'
]


def get_oauth_token():
    registry = getUtility(IRegistry)
    oauth_token = registry.get('plone.google_oauth_token', None)
    if oauth_token is None:
        return
    try:
        return json.loads(oauth_token)
    except TypeError:
        return None


def get_youtube_service():
    oauth_token = get_oauth_token()
    return get_service('youtube', 'v3', SCOPES,
                       access_token=oauth_token['access_token'],
                       refresh_token=oauth_token['refresh_token'])


def should_upload(obj):
    if obj.portal_type != 'Video':
        return False

    if get_oauth_token() is None:
        return False
    value = getattr(obj, 'upload_to_youtube', None)
    if value is None or value in (False, 'false', 'False', 'f', '0'):
        return False
    return True


def upload(obj, filepath, privacy=PRIVATE_VIDEO_STATUS,
           chunksize=DEFAULT_CHUNK_SIZE, filename=None):
    # Should we use private instead of unlisted? I don't know here...

    try:
        service = get_youtube_service()
    except Exception:
        logger.error("Can't get youtube service.", exc_info=True)
        return False

    video_service = service.videos()

    body = dict(
        snippet=dict(
            title=obj.title,
            description=obj.description
        ),
        status=dict(
            privacyStatus=privacy
        )
    )

    mimetype = None
    if filename is not None:
        mimetype, _ = mimetypes.guess_type(filename)

    media_body = MediaFileUpload(
        filepath,
        mimetype=mimetype,
        chunksize=chunksize,
        resumable=True
    )

    insert_request = video_service.insert(
        body=body,
        part=",".join(body.keys()),
        media_body=media_body)

    resumable_upload(obj, service, insert_request)


def resumable_upload(obj, service, insert_request):
    response = None
    retry = 0
    while response is None:
        try:
            status, response = insert_request.next_chunk()
            if 'id' in response:
                # uploaded successfully
                if not getCelery().conf.task_always_eager:
                    obj._p_jar.sync()
                obj.youtube_url = 'https://youtu.be/{id}'.format(
                    id=response['id'])
                if IUploadedToYoutube.providedBy(obj):
                    try:
                        delete(obj._youtube_video_id)
                    except Exception:
                        logger.warning(
                            'Error deleting existing youtube video: {}'.format(
                                obj._youtube_video_id
                            ), exc_info=True)
                obj._youtube_video_id = response['id']
                alsoProvides(obj, IUploadedToYoutube)
                _update_access_token(service)
            else:
                logger.error('Youtube upload failed with an unexpected'
                             ' response: {}'.format(response))
        except Exception:
            logger.warning(
                'An error occured while uploading to youtube {e}', exc_info=True)
            retry += 1
            time.sleep(3)
            if retry > MAX_RETRIES:
                # bubble up error
                raise


def _update_access_token(service):
    registry = getUtility(IRegistry)
    oauth_token = registry.get('plone.google_oauth_token', None)
    if oauth_token is None:
        return
    try:
        token = json.loads(oauth_token)
        creds = service._http.request.credentials
        current = creds.token_response
        if token['access_token'] != current['access_token']:
            current = current.copy()
            if 'refresh_token' not in current:
                current['refresh_token'] = creds.refresh_token
            registry['plone.google_oauth_token'] = json.dumps(
                current).decode('utf-8')
    except (TypeError, AttributeError, KeyError):
        pass


def _get_video(service, obj=None, video_id=None, part='snippet'):
    if obj is not None:
        video_id = obj._youtube_video_id
    video_service = service.videos()
    results = video_service.list(
        part=part,
        id=video_id
    ).execute()

    if len(results['items']) == 0:
        raise Exception('Expected video id {} to be found on youtube account'.format(
            video_id
        ))

    return results['items'][0]


def edit(obj):
    # update tile/description
    try:
        service = get_youtube_service()
    except Exception:
        logger.error("Can't get youtube service.", exc_info=True)
        return False
    video = _get_video(service, obj)
    video_service = service.videos()

    if (video['snippet']['title'] != obj.title or
            video['snippet']['description'] != obj.description):
        video['snippet']['title'] = obj.title
        video['snippet']['description'] = obj.description
        video_service.update(
            part='snippet',
            body=video
        ).execute()
    _update_access_token(service)


def publish(obj):
    try:
        service = get_youtube_service()
    except Exception:
        logger.error("Can't get youtube service.", exc_info=True)
        return False

    video = _get_video(service, obj, part='status')
    video_service = service.videos()
    video['status']['privacyStatus'] = 'public'
    video_service.update(
        part='status',
        body=video
    ).execute()
    _update_access_token(service)


def unlist(obj):
    try:
        service = get_youtube_service()
    except Exception:
        logger.error("Can't get youtube service.", exc_info=True)
        return False

    video = _get_video(service, obj, part='status')
    video_service = service.videos()
    video['status']['privacyStatus'] = PRIVATE_VIDEO_STATUS
    video_service.update(
        part='status',
        body=video
    ).execute()
    _update_access_token(service)


def delete(video_id):
    try:
        service = get_youtube_service()
    except Exception:
        logger.error("Can't get youtube service.", exc_info=True)
        return False
    video_service = service.videos()
    video_service.delete(
        id=video_id, onBehalfOfContentOwner=None).execute()
    _update_access_token(service)



from plone.namedfile.interfaces import INamedImage
from zope.interface import Attribute
from OFS.interfaces import IApplication
from plone.app.contenttypes.interfaces import IFile
from plone.supermodel import model
from Products.CMFPlone.interfaces import IHideFromBreadcrumbs
from zope.interface import Interface
from zope import schema


class ICastleApplication(IApplication):
    pass


class IDashboard(IHideFromBreadcrumbs):
    pass


class IMedia(model.Schema, IFile):
    pass


class IVideo(IMedia):
    pass


class IAudio(IMedia):
    pass


class ISlideshow(Interface):
    custom_dom_id = schema.Text(
        title=u"ID for the slideshow element",
        description=u"If custom styling desired for this slideshow",
        required=False
    )


class ITrashed(Interface):
    """
    marker for object that is in the trash
    """


class IHasDefaultImage(Interface):
    """
    marker interface for content to specify that it has default image
    when a manual image is not specified
    """


class IReferenceNamedImage(INamedImage):
    reference = Attribute('')


class IUploadedToYoutube(Interface):
    """Marker interface for videos that have been uploaded to YouTube"""



from tendo import singleton
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from plone import api
from .utils import setup_site
from castle.cms.pwexpiry.utils import days_since_event
from castle.cms.pwexpiry.utils import send_notification_email
from DateTime import DateTime
import time

import logging

# loop through acl users of each site, making sure they have the pwexpiry related
# data on them and send out expiration emails/warnings


logger = logging.getLogger('castle.cms')


def update_password_expiry(site):
    setup_site(site)

    pwexpiry_enabled = api.portal.get_registry_record('plone.pwexpiry_enabled', default=False)
    validity_period = api.portal.get_registry_record('plone.pwexpiry_validity_period', default=0)

    for user in api.user.get_users():
        if pwexpiry_enabled and validity_period > 0:
            whitelist = api.portal.get_registry_record('plone.pwexpiry_whitelisted_users', default=[])
            whitelisted = whitelist and user.getId() in whitelist
            if not whitelisted:
                password_date = user.getProperty('password_date', None)
                current_time = DateTime()

                if password_date:
                    since_last_pw_reset = days_since_event(
                        password_date.asdatetime(),
                        current_time.asdatetime()
                    )

                    remaining_days = validity_period - since_last_pw_reset
                    if remaining_days < 0:
                        # Password has expired
                        user.setMemberProperties({
                            'reset_password_required': True,
                            'reset_password_time': time.time()
                        })
                        continue
                    if remaining_days < 2:
                        send_notification_email(user, remaining_days)
                else:
                    user.setMemberProperties({
                        'password_date': current_time
                    })


def run(app):
    singleton.SingleInstance('pwexpiry')

    user = app.acl_users.getUser('admin')  # noqa
    newSecurityManager(None, user.__of__(app.acl_users))  # noqa

    for oid in app.objectIds():  # noqa
        obj = app[oid]  # noqa
        if IPloneSiteRoot.providedBy(obj):
            try:
                update_password_expiry(obj)
            except Exception:
                logger.error('Could not update password expiry data for %s' % oid, exc_info=True)


if __name__ == '__main__':
    run(app)  # noqa



from castle.cms.toolbar import BaseToolbarModifier
from collective.easyform.interfaces import IEasyForm
from collective.easyform.interfaces import IEasyFormLayer
from plone import api
from plone.schemaeditor.interfaces import IFieldContext
from plone.schemaeditor.interfaces import ISchemaContext


class ToolbarModifier(BaseToolbarModifier):
    """
    Add menu item to review migration
    """
    layer = IEasyFormLayer

    def __call__(self, result_menu, name, view):
        context = view.real_context
        if (name != 'toolbar_menu' or
                (len(result_menu) and result_menu[0]['id'] != 'view-page') or
                (not IFieldContext.providedBy(context) and
                    not ISchemaContext.providedBy(context))):
            return result_menu

        # find easyform context
        count = 0
        while not IEasyForm.providedBy(context):
            try:
                context = context.__parent__
            except AttributeError:
                context = api.portal.get()
                break
            count += 1
            if count > 20:
                # prevent getting stuck
                context = api.portal.get()
                break

        result_menu[0]['url'] = '{}/view'.format(context.absolute_url())
        return result_menu



from Products.CMFPlone.controlpanel.browser import quickinstaller
from Products.CMFCore.utils import getToolByName
import Products.CMFPlone
import logging
from Products.CMFPlone import PloneMessageFactory as _
from Products.statusmessages.interfaces import IStatusMessage
from plone.memoize import view


class ManageProductsView(quickinstaller.ManageProductsView):
    upgradable_only = (
        'castle.cms',
        'castle.theme',
        'plone.app.standardtiles',
        'plone.app.tiles',
        'plone.app.mosaic',
        'collective.documentviewer',
        'collective.elasticsearch',
        'Products.CMFPlone'
    )

    @view.memoize
    def marshall_addons(self):
        addons = super(ManageProductsView, self).marshall_addons()
        for product_id in ('castle.cms', 'castle.theme', 'plone.app.mosaic'):
            profile_id = '{}:default'.format(product_id)
            profile_version = str(self.ps.getVersionForProfile(profile_id))
            installed_profile_version = str('.'.join(self.ps.getLastVersionForProfile(profile_id)))
            addons[product_id]['upgrade_info'] = dict(
                required=profile_version != installed_profile_version,
                available=len(self.ps.listUpgrades(profile_id)) > 0,
                hasProfile=True,
                installedVersion=installed_profile_version,
                newVersion=profile_version,
            )
        return addons

    def get_installed(self):
        return [p for p in self.get_addons(apply_filter='installed').values()
                if p['id'] not in self.upgradable_only]

    def get_available(self):
        return [p for p in self.get_addons(apply_filter='available').values()
                if p['id'] not in self.upgradable_only]

    def upgrade_product(self, product):
        '''
        Also upgrade profiles
        '''
        # XXX remove customization in future versions!!!
        if not Products.CMFPlone.__version__.startswith('5.0'):
            raise Exception('Incompatible upgrade mechanisms')
        if super(ManageProductsView, self).upgrade_product(product):
            messages = IStatusMessage(self.request)
            try:
                profile_id = '{}:default'.format(product)
                steps_to_run = self.ps.listUpgrades(profile_id)
                if steps_to_run:
                    self.ps.upgradeProfile(profile_id)
                return True
            except Exception as ex:
                logging.error("Could not upgrade %s: %s" % (product, ex))
                messages.addStatusMessage(
                    _(u'Error upgrading ${product}.', mapping={'product': product}), type="error")
                return False
        else:
            return False


class UpgradeProductsView(quickinstaller.UpgradeProductsView):

    def __call__(self):
        qi = ManageProductsView(self.context, self.request)
        products = self.request.get('prefs_reinstallProducts', None)
        if products:
            for product in products:
                qi.upgrade_product(product)

        purl = getToolByName(self.context, 'portal_url')()
        self.request.response.redirect(purl + '/prefs_install_products_form')



from zope.interface import Interface


class IAuthenticator(Interface):

    def __init__(context, request):
        pass

    def get_secure_flow_key(username=None):
        pass

    def get_secure_flow_state(username=None):
        pass

    def set_secure_flow_state(username=None, state='requesting-auth-code'):
        pass

    def change_password(member, new_password):
        pass

    def issue_country_exception_request(user, country):
        pass

    def login(user):
        pass

    def authenticate(username=None, password=None, country=None, login=True):
        pass

    def issue_2factor_code(username):
        pass

    def authorize_2factor(username, code):
        pass

    def get_supported_auth_schemes():
        pass

    def get_options():
        pass



# -*- coding: utf-8 -*-
import json

from castle.cms.interfaces import ISocialMediaSchema
from castle.cms.services.google import GOOGLE_CLIENT_ID
from castle.cms.services.google import GOOGLE_CLIENT_SECRET
from castle.cms.services import twitter
from castle.cms.services.google import youtube
from oauth2client.client import OAuth2WebServerFlow
from plone import api
from plone.app.registry.browser import controlpanel
from plone.protect.interfaces import IDisableCSRFProtection
from plone.registry.interfaces import IRegistry
from Products.CMFPlone import PloneMessageFactory as _
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form.interfaces import HIDDEN_MODE
from zope.component import getUtility
from zope.interface import alsoProvides


class AuthorizeTwitter(BrowserView):

    def __call__(self):
        if 'oauth_verifier' in self.request.form:
            alsoProvides(self.request, IDisableCSRFProtection)
            tokens = twitter.authorize()
            registry = getUtility(IRegistry)
            registry['plone.twitter_oauth_token'] = tokens['oauth_token']
            registry['plone.twitter_oauth_secret'] = tokens['oauth_token_secret']
            registry['plone.twitter_screen_name'] = tokens['screen_name']
            self.request.response.redirect('%s/@@social-controlpanel' % (
                api.portal.get().absolute_url()))
        else:
            self.request.response.redirect(twitter.get_auth_url())


class AuthorizeGoogle(BrowserView):

    def __call__(self):
        client = OAuth2WebServerFlow(
            client_id=GOOGLE_CLIENT_ID,
            client_secret=GOOGLE_CLIENT_SECRET,
            scope=youtube.SCOPES,
            redirect_uri='{}/@@google-auth'.format(
                self.context.absolute_url()))
        if 'code' in self.request.form:
            creds = client.step2_exchange(self.request.form['code'])
            alsoProvides(self.request, IDisableCSRFProtection)
            registry = getUtility(IRegistry)
            registry['plone.google_oauth_token'] = json.dumps(
                creds.token_response).decode('utf-8')
            self.request.response.redirect('%s/@@social-controlpanel' % (
                api.portal.get().absolute_url()))
        else:
            self.request.response.redirect(
                client.step1_get_authorize_url())


class SocialControlPanelForm(controlpanel.RegistryEditForm):

    id = "SocialControlPanel"
    label = _(u"Social Media Settings")
    description = _("Social media sharing settings.")
    schema = ISocialMediaSchema
    schema_prefix = "plone"

    def updateWidgets(self):
        super(SocialControlPanelForm, self).updateWidgets()
        self.widgets["twitter_oauth_token"].mode = HIDDEN_MODE
        self.widgets["twitter_oauth_secret"].mode = HIDDEN_MODE
        self.widgets["twitter_screen_name"].mode = HIDDEN_MODE
        self.widgets["google_oauth_token"].mode = HIDDEN_MODE


class SocialControlPanel(controlpanel.ControlPanelFormWrapper):
    form = SocialControlPanelForm

    index = ViewPageTemplateFile('templates/social.pt')

    def get_auths(self):
        keys = twitter.get_keys()
        registry = getUtility(IRegistry)
        google_auth = True
        check_keys = ('plone.google_oauth_token',)
        for key in check_keys:
            if not registry.get(key, None):
                google_auth = False

        return {
            'twitter': {
                'client': keys[0] is not None and keys[1] is not None,
                'authorized': twitter.get_auth() is not None
            },
            'google': {
                'client': (GOOGLE_CLIENT_ID is not None and
                           GOOGLE_CLIENT_SECRET is not None),
                'authorized': google_auth
            }
        }



from .api import verify_recaptcha  # noqa
from .catalog import ESConnectionFactoryFactory  # noqa
from .catalog import add_indexes  # noqa
from .catalog import add_metadata  # noqa
from .catalog import delete_indexes  # noqa
from .catalog import delete_metadata  # noqa
from .catalog import index_in_es  # noqa
from .content import clear_object_cache  # noqa
from .content import get_context_from_request  # noqa
from .content import get_data_from_url  # noqa
from .content import get_folder_contents  # noqa
from .content import get_image_info  # noqa
from .content import get_object_version  # noqa
from .content import get_paste_data  # noqa
from .content import get_path  # noqa
from .content import has_image  # noqa
from .content import inline_images_in_dom  # noqa
from .content import is_max_paste_items  # noqa
from .content import is_mosaic_edit_form  # noqa
from .content import recursive_create_path  # noqa
from .mail import get_email_from_address  # noqa
from .mail import send_email  # noqa
from .misc import get_ip  # noqa
from .misc import get_random_string  # noqa
from .misc import json_dumps  # noqa
from .misc import make_random_key  # noqa
from .misc import normalize_url  # noqa
from .misc import retriable  # noqa
from .misc import strings_differ  # noqa
from .query import parse_query_from_data  # noqa
from .registry import get_backend_url  # noqa
from .registry import get_chat_info  # noqa
from .registry import get_public_url  # noqa
from .registry import get_upload_fields  # noqa
from .registry import no_backend_url  # noqa
from .registry import site_has_icon  # noqa
from .security import get_managers  # noqa
from .security import get_permission_title  # noqa
from .security import publish_content  # noqa
from .security import is_backend  # noqa
from .text import truncate_text  # noqa
from .text import truncate_text as truncateText  # noqa



from plone.app.mosaic.interfaces import IMosaicRegistryAdapter
from Products.CMFPlone.interfaces import ITinyMCESchema
from zope.component import getUtility
from castle.cms.tiles.dynamic import get_tile_manager
from plone.registry.interfaces import IRegistry
from plone.app.mosaic.registry import MosaicRegistry
from plone import api
from zope.interface import implementer
from zope.component import adapter
from castle.cms import cache


MOSAIC_CACHE_DURATION = 1 * 60
_rich_text_widget_types = (
    'plone_app_z3cform_wysiwyg_widget_WysiwygWidget',
    'plone_app_z3cform_wysiwyg_widget_WysiwygFieldWidget',
    'plone_app_widgets_dx_RichTextWidget',
    'plone_app_z3cform_widget_RichTextFieldWidget',
)


@implementer(IMosaicRegistryAdapter)
@adapter(IRegistry)
class CastleMosaicRegistry(MosaicRegistry):
    """
    Overrides mosaic registry adapter
    """
    def parseRegistry(self):
        cache_key = '%s-mosaic-registry' % '/'.join(
            api.portal.get().getPhysicalPath()[1:])
        if not api.env.debug_mode():
            try:
                return cache.get(cache_key)
            except KeyError:
                result = super(CastleMosaicRegistry, self).parseRegistry()
        else:
            result = super(CastleMosaicRegistry, self).parseRegistry()

        mng = get_tile_manager()
        for tile in mng.get_tiles():
            if tile.get('hidden'):
                continue
            key = 'castle_cms_dynamic_{}'.format(tile['id'])
            category = tile.get('category') or 'advanced'
            category_id = category.replace(' ', '_').lower()
            if category_id not in result['plone']['app']['mosaic']['tiles_categories']:
                result['plone']['app']['mosaic']['tiles_categories'][category_id] = {
                    'label': category,
                    'name': category_id,
                    'weight': 100
                }
            result['plone']['app']['mosaic']['app_tiles'][key] = {
                'category': category_id,
                'default_value': None,
                'favorite': False,
                'label': tile['title'],
                'name': tile['name'],
                'tile_type_id': u'castle.cms.dynamic',
                'read_only': False,
                'rich_text': False,
                'settings': True,
                'tile_type': u'app',
                'weight': tile['weight']
            }

        registry = getUtility(IRegistry)
        settings = registry.forInterface(
            ITinyMCESchema, prefix="plone", check=False)
        if settings.libraries_spellchecker_choice != 'AtD':
            cache.set(cache_key, result, MOSAIC_CACHE_DURATION)
            return result

        # add atd config to toolbar dynamically
        mos_settings = result['plone']['app']['mosaic']
        mos_settings['richtext_toolbar']['AtD'] = {
            'category': u'actions',
            'name': u'toolbar-AtD',
            'weight': 0,
            'favorite': False,
            'label': u'After the deadline',
            'action': u'AtD',
            'icon': False
        }
        for widget_type in _rich_text_widget_types:
            mos_settings['widget_actions'][widget_type]['actions'].append('toolbar-AtD')  # noqa
        mos_settings['structure_tiles']['text']['available_actions'].append('toolbar-AtD')  # noqa
        mos_settings['app_tiles']['plone_app_standardtiles_rawhtml']['available_actions'].append('toolbar-AtD')  # noqa
        cache.set(cache_key, result, MOSAIC_CACHE_DURATION)
        return result



from plone.registry.interfaces import IRegistry
from Products.Five import BrowserView
from zope.component import getUtility
import json


class FirstVisitMessage(BrowserView):
    def __call__(self):
        registry = getUtility(IRegistry)
        displayMsg = registry.get("castle.show_disclaimer")

        self.request.response.setHeader('Content-type', 'application/json')

        result = {
            "enabled": displayMsg
        }

        if displayMsg:
            result["msg"] = registry.get("castle.site_disclaimer")

        return json.dumps(result)



import re

from lxml.html import fromstring
from lxml.html import tostring
from lxml.html.clean import Cleaner


_truncate_cleaner = Cleaner(
    scripts=True, javascript=True, comments=True, style=True, links=True,
    meta=True, page_structure=True, embedded=True, frames=True, forms=True,
    annoying_tags=True, remove_tags=('div',), kill_tags=('img', 'hr'),
    remove_unknown_tags=True)


def truncate_text(text, max_words=30, more_link=None, clean=False):
    """
    adapted from Django
    """

    if not isinstance(text, basestring):
        return ''

    if clean:
        try:
            if not isinstance(text, unicode):
                text = text.decode('utf8')
            xml = fromstring(text)
            _truncate_cleaner(xml)
            # also remove empty tags...
            for p in xml.xpath("//p"):
                if len(p):
                    continue
                t = p.text
                if not (t and t.replace('&nbsp;', '').strip()):
                    p.getparent().remove(p)
            text = tostring(xml)
        except Exception:
            pass
    length = int(max_words)
    if length <= 0:
        return u''
    html4_singlets = ('br', 'col', 'link', 'base',
                      'img', 'param', 'area', 'hr', 'input')
    # Set up regular expressions
    re_words = re.compile(r'&.*?;|<.*?>|(\w[\w-]*)', re.U)
    re_tag = re.compile(r'<(/)?([^ ]+?)(?: (/)| .*?)?>')
    # Count non-HTML words and keep note of open tags
    pos = 0
    end_text_pos = 0
    words = 0
    open_tags = []
    while words <= length:
        m = re_words.search(text, pos)
        if not m:
            # Checked through whole string
            break
        pos = m.end(0)
        if m.group(1):
            # It's an actual non-HTML word
            words += 1
            if words == length:
                end_text_pos = pos
            continue
        # Check for tag
        tag = re_tag.match(m.group(0))
        if not tag or end_text_pos:
            # Don't worry about non tags or tags after our truncate point
            continue
        closing_tag, tagname, self_closing = tag.groups()
        tagname = tagname.lower()  # Element names are always case-insensitive
        if self_closing or tagname in html4_singlets:
            pass
        elif closing_tag:
            # Check for match in open tags list
            try:
                i = open_tags.index(tagname)
            except ValueError:
                pass
            else:
                # SGML: An end tag closes, back to the matching start tag,
                # all unclosed intervening start tags with omitted end tags
                open_tags = open_tags[i + 1:]
        else:
            # Add it to the start of the open tags list
            open_tags.insert(0, tagname)

    if words <= length:
        # Don't try to close tags if we don't need to truncate
        return text
    out = text[:end_text_pos]
    out += '&hellip;'
    if more_link:
        out += ' <a href="%s">more &#x2192;</a>' % more_link
    # Close any tags still open
    for tag in open_tags:
        out += '</%s>' % tag
    # Return string
    return out



import json

from castle.cms.behaviors.location import ILocation
from castle.cms.behaviors.search import ISearch
from castle.cms.interfaces import ILDData
from castle.cms.utils import site_has_icon
from plone.app.layout.globals.interfaces import IViewView
from plone.tiles import Tile
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFPlone.defaultpage import get_default_page
from Products.CMFPlone.log import logger
from unidecode import unidecode
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.viewlet.interfaces import IViewlet
from zope.viewlet.interfaces import IViewletManager


head_viewlets = {
    'plone.htmlhead.title': 'plone.htmlhead',
    'plone.links.author': 'plone.htmlhead.links',
    'plone.links.RSS': 'plone.htmlhead.links',
    'plone.htmlhead.dublincore': 'plone.htmlhead',
    'plone.htmlhead.socialtags': 'plone.htmlhead'
}


def _date(obj, date):
    try:
        return getattr(obj, date)().ISO8601()
    except Exception:
        return ''


class MetaDataTile(Tile):

    def get_navigation_links(self):
        return u"""
<link rel="home"
      title="Front page"
      href="{url}"/>
<link rel="contents"
      href="{url}/sitemap"/>""".format(url=self.root_url)

    def get_icons(self):
        if not site_has_icon():
            return ''
        return '''
<link rel="apple-touch-icon" sizes="180x180" href="{url}/site-icon.png">
<link rel="icon" type="image/png" href="{url}/site-icon.png?scale=32" sizes="32x32">
<link rel="icon" type="image/png" href="{url}/site-icon.png?scale=16" sizes="16x16">
<link rel="manifest" href="{url}/manifest.json">
<meta name="theme-color" content="#ffffff">'''.format(url=self.root_url)

    def _wrap_ld(self, data):
        return '<script type="application/ld+json">' + json.dumps(data) + '</script>'

    def get_ld_data(self):
        result = ''
        ld = ILDData(self.context, None)
        if ld is None:
            return ''
        result += self._wrap_ld(ld.get_data())

        if ISiteRoot.providedBy(self.context):
            try:
                page = self.context[get_default_page(self.context)]
                result += self._wrap_ld(ILDData(page).get_data())
            except AttributeError:
                pass

        return result

    def get_basic_tags(self):
        try:
            context = self.context
            if ISiteRoot.providedBy(context):
                try:
                    context = context[get_default_page(context)]
                except AttributeError:
                    pass
            tags = {
                'modificationDate': _date(context, 'modified'),
                'publicationDate': _date(context, 'effective'),
                'expirationDate': _date(context, 'expires'),
                'generator': 'CastleCMS 2.5.17',
                "distribution": "Global",
                "robots": "index,follow"
            }
            ldata = ILocation(context, None)
            if ldata is not None:
                if ldata.locations:
                    location = ldata.locations
                    if type(location) in (list, tuple, set):
                        location = location[0]
                    tags['location'] = location

            search = ISearch(context, None)
            if search is not None:
                if search.robot_configuration:
                    config = search.robot_configuration[:]
                    if 'index' not in config:
                        config.append('noindex')
                    if 'follow' not in config:
                        config.append('nofollow')
                    tags['robots'] = ','.join(config)

            return ''.join([u'<meta name="{}" content="{}">'.format(name, value)
                            for name, value in tags.items()])
        except Exception:
            return u''

    def get_search_link(self):
        return '''
<link rel="search"
      title="Search this site"
      href="{url}/@@search" />'''.format(
            url=self.root_url
          )

    def get_printcss_link(self):
        return ''' <link rel="stylesheet" href="{url}/++plone++castle/less/public/print.css" type="text/css" media="print">'''.format(  # noqa
              url=self.root_url
              )

    def get_canonical_url(self):
        context_state = getMultiAdapter(
            (self.context, self.request), name=u'plone_context_state')
        canonical_object = context_state.canonical_object()
        if self.context != canonical_object:
            can_state = getMultiAdapter(
                (canonical_object, self.request), name=u'plone_context_state')
            url = can_state.view_url()
        else:
            url = context_state.view_url()
        return u'    <link rel="canonical" href="%s" />' % url

    def __call__(self):
        portal_state = getMultiAdapter((self.context, self.request),
                                       name=u'plone_portal_state')
        self.root_url = portal_state.navigation_root_url()

        result = self.get_canonical_url()
        alsoProvides(self, IViewView)
        for name, manager_name in head_viewlets.items():
            manager = queryMultiAdapter(
                (self.context, self.request, self),
                IViewletManager, name=manager_name
            )
            viewlet = queryMultiAdapter(
                (self.context, self.request, self, manager),
                IViewlet, name=name
            )
            if viewlet is not None:
                try:
                    viewlet.update()
                    result += viewlet.render()
                except Exception:
                    logger.warn('Error rendering head viewlet %s' % name, exc_info=True)
        result += unidecode(self.get_basic_tags())
        result += unidecode(self.get_navigation_links())
        result += unidecode(self.get_ld_data())
        result += unidecode(self.get_icons())
        result += unidecode(self.get_search_link())
        result += unidecode(self.get_printcss_link())
        return u'<html><head>%s</head></html>' % result



from castle.cms.cron._pw_expiry import update_password_expiry
from plone.registry.interfaces import IRegistry
from plone.registry import field
from plone.registry import Record
from plone import api
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility

PROFILE_ID = 'profile-castle.cms:2_5_17'


def upgrade(context, logger=None):
    setup = getToolByName(context, 'portal_setup')
    setup.runAllImportStepsFromProfile(PROFILE_ID)
    registry = getUtility(IRegistry)
    try:
        auth_step_timeout = api.portal.get_registry_record(name='plone.auth_step_timeout', default=None)
    except api.exc.InvalidParameterError:
        auth_step_timeout = None

    if not auth_step_timeout:
        auth_timeout_field = field.Int(title=u"(Seconds) This amount of inactivity \
                                                will reset the login process",
                                       description=u"Between each step, the allowed \
                                                     time is reset to this amount",
                                       default=120)
        auth_timeout_record = Record(auth_timeout_field)
        registry._records['plone.auth_step_timeout'] = auth_timeout_record

    try:
        disclaimer = api.portal.get_registry_record(name='castle.show_disclaimer', default=None)
    except api.exc.InvalidParameterError:
        disclaimer = None

    if not disclaimer:
        show_disclaimer_field = field.Bool(
            title=u'Show disclaimer for first time a user visits a site. '
                  u'To comply with ePrivacy Directive, use this feature to notify about cookie use.',
            default=False,
            required=False)
        site_disclaimer_field = field.Text(
            title=u"Disclaimer",
            default=u'<p><strong>Disclaimer</strong> '
                    u'<em>You are seeing this because this is your first time visiting the site.</em></p>',
            required=False)
        show_disclaimer_record = Record(show_disclaimer_field)
        site_disclaimer_record = Record(site_disclaimer_field)
        registry._records['castle.show_disclaimer'] = show_disclaimer_record
        registry._records['castle.site_disclaimer'] = site_disclaimer_record

    update_password_expiry(api.portal.get())

    for id in context.objectIds():
        context[id].reindexObject(idxs=['has_private_parents'])

    try:
        scrub_site_identity = api.portal.get_registry_record(name='plone.scrub_title_logo_to_backend_login',
                                                             default=None)
    except api.exc.InvalidParameterError:
        scrub_site_identity = None

    if not scrub_site_identity:
        scrub_identity_field = field.Bool(
            title=u'Do not show identifying customizations (logo/text) at backend login view',
            description=u'If set, backend visitors will see a generic login view.',
            default=False,
            required=False)
        scrub_identity_record = Record(scrub_identity_field)
        registry._records['plone.scrub_title_logo_to_backend_login'] = scrub_identity_record

    try:
        login_footer_message = api.portal.get_registry_record(name='plone.scrub_title_logo_to_backend_login',
                                                              default=None)
    except api.exc.InvalidParameterError:
        login_footer_message = None

    if not login_footer_message:
        login_footer_message_field = field.TextLine(
            title=u'Login Footer Message',
            description=u'Display a message or warning below the secure login form',
            required=False)
        login_footer_message_record = Record(login_footer_message_field)
        registry._records['plone.login_footer_message'] = login_footer_message_record



# -*- coding: utf-8 -*-
import pkg_resources

from plone.app.dexterity.browser import fields
from plone.schemaeditor.browser.schema.listing import ReadOnlySchemaListing
from zope.event import notify
from z3c.form.events import DataExtractedEvent


class EnhancedSchemaListing(fields.EnhancedSchemaListing):
    def extractData(self, setErrors=True):
        data, errors = super(EnhancedSchemaListing, self).extractData(setErrors=setErrors)
        for group in self.groups:
            groupData, groupErrors = group.extractData(setErrors=setErrors)
            data.update(groupData)
            if groupErrors:
                if errors:
                    errors += groupErrors
                else:
                    errors = groupErrors
        for fname, value in data.items():
            if fname not in self.context.schema:
                del data[fname]
        notify(DataExtractedEvent(data, errors, self))
        return data, errors


if pkg_resources.get_distribution('plone.resourceeditor'):
    advancedbtns = EnhancedSchemaListing.buttons.copy()
    regularbtns = EnhancedSchemaListing.buttons.copy().omit('modeleditor')


class TypeFieldsPage(fields.TypeFieldsPage):
    def __init__(self, context, request):
        super(TypeFieldsPage, self).__init__(context, request)
        if pkg_resources.get_distribution('plone.resourceeditor'):
            if 'castlecms-advanced=1' in request['HTTP_COOKIE']:
                EnhancedSchemaListing.buttons = advancedbtns
            else:
                EnhancedSchemaListing.buttons = regularbtns

    @property
    def form(self):
        if self.context.fti.hasDynamicSchema:
            return EnhancedSchemaListing
        else:
            return ReadOnlySchemaListing



import json
import logging
import time
from multiprocessing import Pool
from urllib import urlencode

import requests

import transaction
from AccessControl.SecurityManagement import newSecurityManager
from Acquisition import aq_parent
from BTrees.OOBTree import OOBTree
from castle.cms.social import COUNT_ANNOTATION_KEY
from castle.cms.utils import clear_object_cache, index_in_es, retriable
from plone.app.redirector.interfaces import IRedirectionStorage
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import defaultpage
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from tendo import singleton
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility
from zope.component.hooks import setSite


USE_MULTIPROCESSING = True

MATOMO_TOKEN_AUTH = 'castle.matomo_token_auth'
MATOMO_BASE_URL = 'castle.matomo_base_url'
MATOMO_SITE_ID = 'castle.matomo_site_id'

logger = logging.getLogger('castle.cms')


def get_matomo_api_url(url):
    registry = getUtility(IRegistry)
    site_id = registry.get(MATOMO_SITE_ID, None)
    base_url = registry.get(MATOMO_BASE_URL, None)
    token_auth = registry.get(MATOMO_TOKEN_AUTH, None)
    params = {
        'module': 'API',
        'method': 'Actions.getOutlinks',
        'idSite': site_id,
        'period': 'year',
        'date': 'today',
        'format': 'json',
        'token_auth': token_auth,
        'expanded': '1',
        'filter_pattern_recursive': '',
        'u': url
    }
    return '{}/?{}'.format(
        base_url,
        urlencode(params)
    )


_matomo_data_mapping = {
    'facebook_matomo': 'www.facebook.com',
    'twitter_matomo': 'twitter.com'
}


def get_matomo_url_data(urls):
    data = {
        'facebook_matomo': 0,
        'twitter_matomo': 0
    }
    for url in urls:
        query_url = get_matomo_api_url(url)
        resp = requests.get(query_url, timeout=10).content
        datatable = json.loads(resp)
        for d in datatable:
            for key, label in _matomo_data_mapping.items():
                if d['label'] == label:
                    data[key] += d['nb_visits']
    return data


COUNT_TYPES = {
    'pinterest': {
        'url': 'http://api.pinterest.com/v1/urls/count.json?callback=foobar&url=%s',
        'slash_matters': True
    },
    'matomo': {
        'generator': get_matomo_url_data,
    }
}


def _get_url_data(args):
    config, url = args
    access_url = config['url'] % url
    resp = requests.get(access_url).content.lstrip('foobar(').rstrip(')')
    try:
        data = json.loads(resp)
    except ValueError:
        return 0
    if 'count' in data:
        return data['count']
    else:
        return 0


_req_pool = Pool(6)


def _get_urls_data(args):
    results = []
    config, urls = args
    if config.get('generator'):
        results.append(config['generator'](urls))
    else:
        req_urls = []
        for orig_url in urls:
            req_urls.append((config, orig_url))
            if config.get('slash_matters'):
                req_urls.append((config, orig_url.rstrip('/') + '/'))
        if USE_MULTIPROCESSING:
            results = _req_pool.map(_get_url_data, req_urls)
        else:
            for req_url in req_urls:
                results.append(_get_url_data(req_url))
    return results


_pool = Pool(processes=3)


def _get_counts(urls, count_types):
    counts = {}
    for type_, config in count_types.items():
        for result in _get_urls_data((config, urls)):
            if isinstance(result, dict):
                # manually setting keys on social data here
                for key, value in result.items():
                    if key not in counts:
                        counts[key] = 0
                    counts[key] += value
            else:
                if type_ not in counts:
                    counts[type_] = 0
                counts[type_] += result
    return counts


def _merge_counts(one, two):
    for key, count in two.items():
        if key in one:
            one[key] += count
        else:
            one[key] = count
    return one


def _has_data(data):
    found = False
    for key, val in data.items():
        if val > 0:
            found = True
            break
    return found


def _count_diff(existing, new):
    diff = False
    for name, value in new.items():
        if value != existing.get(name):
            diff = True
            break
    return diff


@retriable(sync=True)
def get_social_counts(site, obj, site_url, count_types, count=0):
    counts = {}
    site_path = '/'.join(site.getPhysicalPath())
    obj_path = '/'.join(obj.getPhysicalPath())
    rel_path = obj_path[len(site_path):].strip('/')
    print('Looking up ' + obj_path)

    urls = [site_url.rstrip('/') + '/' + rel_path]
    registry = getUtility(IRegistry)
    if obj.portal_type in registry.get('plone.types_use_view_action_in_listings', []):
        urls.append(urls[0] + '/view')

    container = aq_parent(obj)
    if defaultpage.is_default_page(container, obj):
        container_path = '/'.join(container.getPhysicalPath())
        rel_path = container_path[len(site_path):].strip('/')
        urls.append(site_url.rstrip('/') + '/' + rel_path)

    redirector = getUtility(IRedirectionStorage)
    for redirect in redirector.redirects(obj_path):
        rel_path = redirect[len(site_path):].strip('/')
        urls.append(site_url.rstrip('/') + '/' + rel_path)

    counts = _get_counts(urls, count_types)

    if not _has_data(counts):
        return

    obj._p_jar.sync()
    annotations = IAnnotations(obj)
    existing = annotations.get(COUNT_ANNOTATION_KEY, OOBTree())

    if not _count_diff(existing, counts):
        return

    # XXX check if value different first before transaction!
    existing.update(counts)
    annotations[COUNT_ANNOTATION_KEY] = existing

    transaction.commit()

    index_in_es(obj)
    if count % 200 == 0:
        clear_object_cache(site)


def retrieve(site):
    setSite(site)
    registry = getUtility(IRegistry)
    site_url = registry.get('plone.public_url')
    if not site_url:
        logger.info("No public URL is set; skipping site %s" % site)
        return

    # Which counting strategies should we use?
    count_types = COUNT_TYPES.copy()
    site_id = registry.get(MATOMO_SITE_ID, None)
    base_url = registry.get(MATOMO_BASE_URL, None)
    token_auth = registry.get(MATOMO_TOKEN_AUTH, None)
    if not site_id or not base_url or not token_auth:
        # filter all _matomo count types
        for ctype in list(count_types.keys()):
            if ctype.endswith('_matomo'):
                del count_types[ctype]

    catalog = getToolByName(site, 'portal_catalog')
    count = 0
    for brain in catalog(review_state='published'):
        # we ignore some types...
        if brain.portal_type in ('Image', 'Dashboard',):
            continue
        path = brain.getPath()
        count += 1
        try:
            obj = brain.getObject()
            get_social_counts(site, obj, site_url, count_types, count)
            logger.info('retrieved social stats for: %s' % path)
        except Exception:
            logger.warn('error getting social count totals for: %s' % path,
                        exc_info=True)
        time.sleep(2)


def run(app):
    singleton.SingleInstance('socialcounts')

    user = app.acl_users.getUser('admin')  # noqa
    newSecurityManager(None, user.__of__(app.acl_users))  # noqa

    for oid in app.objectIds():  # noqa
        obj = app[oid]  # noqa
        if IPloneSiteRoot.providedBy(obj):
            retrieve(obj)


if __name__ == '__main__':
    run(app)  # noqa



import requests
from plone.registry.interfaces import IRegistry
from zope.component import getUtility
from zope.globalrequest import getRequest


def verify_recaptcha(req=None):
    if req is None:
        req = getRequest()

    registry = getUtility(IRegistry)
    key = registry.get('castle.recaptcha_private_key', '')
    if not key:
        # do not bother verifying if no key is defined
        return True
    code = req.form.get('g-recaptcha-response', '')
    remote_addr = req.get(
        'HTTP_X_FORWARDED_FOR',
        ''
    ).split(',')[0]
    if not remote_addr:
        remote_addr = req.get('REMOTE_ADDR')

    resp = requests.post(
        'https://www.google.com/recaptcha/api/siteverify',
        data=dict(
            secret=key,
            response=code,
            remoteip=remote_addr
        )
    )
    try:
        return resp.json()['success']
    except Exception:
        return False



from castle.cms import utils
from castle.cms.interfaces import ICastleLayer
from castle.cms.interfaces import IGlobalTile
from castle.cms.interfaces import IMetaTile
from castle.cms.interfaces import IVersionViewLayer
from plone import api
from plone.app.standardtiles.contentlisting import ContentListingTile
from plone.resource.interfaces import IResourceDirectory
from plone.tiles.interfaces import IPersistentTile
from plone.tiles.interfaces import ITile
from plone.tiles.interfaces import ITileDataContext
from zope.annotation.interfaces import IAnnotations
from zope.component import adapter
from zope.interface import Interface
from zope.interface import classImplements
from zope.interface import implementer


# make ContentListing persistent
classImplements(ContentListingTile, IPersistentTile)


@implementer(ITileDataContext)
@adapter(Interface, ICastleLayer, IGlobalTile)
def GlobalTileDataContext(context, request, tile):
    """
    Make sure to not pull from drafts data proxy
    """
    return tile.context


@implementer(ITileDataContext)
@adapter(Interface, ICastleLayer, IMetaTile)
def MetaTileDataContext(context, request, tile):
    """
    Make sure to not pull from drafts data proxy
    """
    dc = getattr(tile, 'data_context', None)
    return dc or tile.context


@implementer(ITileDataContext)
@adapter(Interface, IVersionViewLayer, ITile)
def VersionTileDataContext(context, request, tile):
    """
    Get version of context.
    This is used when viewing different versions of content
    """
    version = request.form.get('version')
    return utils.get_object_version(tile.context, version)


@implementer(IAnnotations)
@adapter(IResourceDirectory)
def ResourceDirectoryAnnotations(context):
    # resource directory does not and should not implement annotation
    # storage. We should just proxy to the annotations on the site here
    # This fixes some errors when previewing the theme from the layout editor
    return IAnnotations(api.portal.get())



from zope import schema
from zope.interface import Interface
from plone.autoform import directives as form
from castle.cms.tiles.base import BaseTile
from castle.cms.widgets import RelatedItemFieldWidget
from plone.app.uuid.utils import uuidToCatalogBrain


class StickyFooterTile(BaseTile):
    @property
    def internal_link_url(self):
        if self.data.get('internal_link'):
            try:
                brain = uuidToCatalogBrain(self.data['internal_link'][0])
                return brain.getURL()
            except Exception:
                pass


class IStickyFooterTileSchema(Interface):

    email_text = schema.TextLine(
        title=u"Footer Title",
        description=u"(appears next to email icon)",
        required=False,
        default=u"",
    )

    description = schema.TextLine(
        title=u"Footer Description",
        description=u"(appears left of button)",
        required=False,
        default=u"",
    )

    button_text = schema.TextLine(
        title=u"Footer button text",
        required=False,
        default=u"Subscribe",
    )

    form.widget(internal_link=RelatedItemFieldWidget)
    internal_link = schema.List(
        title=u"Internal Link",
        description=u"Content that the sticky footer button links to",
        required=False,
        default=[],
        value_type=schema.Choice(
            vocabulary="plone.app.vocabularies.Catalog"
        ),
    )

    external_link = schema.TextLine(
        title=u"External URL",
        description=u"If no Interal Link selected, the footer button will link to this URL",
        required=False,
        default=u"",
    )



import json
from castle.cms.tests.utils import get_tile
import logging
import os
import unittest

import requests

import transaction
from castle.cms.browser.search import SearchAjax
from collective.elasticsearch.es import ElasticSearchCatalog
from castle.cms.social import COUNT_ANNOTATION_KEY
from castle.cms.testing import CASTLE_PLONE_INTEGRATION_TESTING
from collective.elasticsearch.interfaces import IElasticSettings
from plone import api
from plone.app.testing import (TEST_USER_ID, TEST_USER_NAME, login,
                               setRoles)
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility

logger = logging.getLogger(__name__)


ES_ENABLED = False
if 'ES_HOST' not in os.environ:
    os.environ['ES_HOST'] = '127.0.0.1'
    logger.warning('ES_HOST not specified in os.environ, using default %s' % str(os.environ['ES_HOST']))

if 'ES_HOST' in os.environ:
    host = os.environ['ES_HOST']
    url = 'http://{}:9200'.format(host)
    try:
        resp = requests.get(url)
        data = resp.json()
        version = data['version']['number']
        if version in ('2.3.2', '2.3.3', '2.3.5'):
            ES_ENABLED = True
        else:
            logger.warning('Unsupported ES version: {}'.format(version))
    except Exception:
        logger.warning('Could not connect to ES: {}'.format(host))


if ES_ENABLED:
    class TestES(unittest.TestCase):

        layer = CASTLE_PLONE_INTEGRATION_TESTING

        def setUp(self):
            self.portal = self.layer['portal']
            self.request = self.layer['request']
            login(self.portal, TEST_USER_NAME)
            setRoles(self.portal, TEST_USER_ID, ('Member', 'Manager'))
            registry = getUtility(IRegistry)
            settings = registry.forInterface(IElasticSettings)
            settings.enabled = True
            settings.sniffer_timeout = None
            catalog = getToolByName(self.portal, 'portal_catalog')
            catalog._elasticcustomindex = 'plone-test-index'
            es = ElasticSearchCatalog(catalog)
            es.recreateCatalog()
            catalog.manage_catalogRebuild()
            transaction.commit()

            # have to do commit for es integration...
            folder = api.content.create(
                type='Folder',
                id='esfolder1',
                container=self.portal,
                title='Foobar folder')
            api.content.create(
                type='Document',
                id='esdoc1',
                container=folder,
                title='Foobar one')
            doc = api.content.create(
                type='Document',
                id='esdoc2',
                container=folder,
                subject=('foobar',),
                title='Foobar two')
            api.content.create(
                type='Document',
                id='esdoc3',
                container=folder,
                title='Foobar three')

            ann = IAnnotations(doc)
            ann[COUNT_ANNOTATION_KEY] = {
                'twitter_matomo': 5,
                'facebook': 5,
            }
            doc.reindexObject()
            transaction.commit()

            url = 'http://{}:9200/plone-test-index/_flush'.format(host)
            requests.post(url)

        def tearDown(self):
            transaction.begin()
            api.content.delete(self.portal.esfolder1)
            transaction.commit()

        def _test_ajax_search_rank_social(self):
            self.request.form.update({
                'SearchableText': 'Foobar',
                'portal_type': 'Document'
            })
            view = SearchAjax(self.portal, self.request)
            result = json.loads(view())
            self.assertEquals(result['count'], 3)
            self.assertEquals(result['results'][0]['path'], '/esfolder1/esdoc2')

        def test_ajax_search_pt(self):
            self.request.form.update({
                'SearchableText': 'Foobar',
                'portal_type': 'Folder'
            })
            view = SearchAjax(self.portal, self.request)
            result = json.loads(view())
            self.assertEquals(result['count'], 1)
            self.assertEquals(result['results'][0]['path'], '/esfolder1')

        def test_ajax_search_subject(self):
            self.request.form.update({
                'SearchableText': 'Foobar',
                'Subject': 'foobar'
            })
            view = SearchAjax(self.portal, self.request)
            result = json.loads(view())
            self.assertEquals(result['count'], 1)
            self.assertEquals(result['results'][0]['path'], '/esfolder1/esdoc2')

        def test_es_querylisting_unicode_issue(self):
            tile = get_tile(self.request, self.portal, 'castle.cms.querylisting', {})
            # should not cause errors...
            self.request.form.update({
                'Title': 'ma\xf1on'
            })
            self.assertTrue(tile.filter_pattern_config != '{}')
            tile()

else:
    class TestEmptyES(unittest.TestCase):
        '''
        test runner throws error if no tests defined in module
        '''

        def test_dummy(self):
            '''
            '''



import json
from plone.app.textfield import RichText
import logging
import time

from castle.cms.tiles.base import BaseTile
from castle.cms.widgets import ImageRelatedItemFieldWidget
from plone import api
from plone.app.theming.interfaces import THEME_RESOURCE_NAME
from plone.app.theming.utils import getCurrentTheme
from plone.app.tiles import MessageFactory as _
from plone.app.tiles.browser import add
from plone.app.tiles.browser import edit
from plone.app.tiles.browser import traversal
from plone.autoform import directives
from plone.autoform.interfaces import WIDGETS_KEY
from plone.memoize import forever
from plone.registry.interfaces import IRegistry
from plone.resource.utils import queryResourceDirectory
from plone.supermodel import model
from plone.supermodel.model import SchemaClass
from plone.tiles.interfaces import IPersistentTile
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from zExceptions import NotFound
from zope import schema
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component import queryUtility
from zope.filerepresentation.interfaces import IRawReadFile
from zope.globalrequest import getRequest
from zope.interface import Invalid
from zope.interface import implements
from zope.schema.vocabulary import SimpleVocabulary
from castle.cms.widgets import RelatedItemFieldWidget, ImageRelatedItemsFieldWidget, RelatedItemsFieldWidget
from collections import OrderedDict

logger = logging.getLogger('castle.cms')


CACHE_KEY = 'castle.cms.tiles.dynamic'


def ChoiceFieldFactory(**options):
    if 'vocabulary' not in options:
        options['vocabulary'] = SimpleVocabulary.fromValues([])
    else:
        options['vocabulary'] = SimpleVocabulary.fromValues(options['vocabulary'])
    return schema.Choice(**options)


def validate_image(val):
    if val and len(val) != 1:
        raise Invalid("Must select 1 image")
    if val:
        utils = getMultiAdapter((api.portal.get(), getRequest()),
                                name="castle-utils")
        obj = utils.get_object(val[0])
        if not obj or obj.portal_type != 'Image':
            raise Invalid('Must provide image file')
    return True


def validate_content(val):
    if val and len(val) != 1:
        raise Invalid("Must select 1 item")
    return True


FIELD_TYPE_MAPPING = {
    'text': schema.TextLine,
    'richtext': {
        'factory': RichText,
        'options': {
            'default_mime_type': 'text/html',
            'output_mime_type': 'text/html',
            'allowed_mime_types': ('text/html',),
        }
    },
    'int': schema.Int,
    'float': schema.Float,
    'decimal': schema.Decimal,
    'datetime': schema.Datetime,
    'date': schema.Date,
    'timedelta': schema.Timedelta,
    'time': schema.Time,
    'password': schema.Password,
    'boolean': schema.Bool,
    'choice': ChoiceFieldFactory,
    'uri': schema.URI,
    'dottedname': schema.DottedName,
    'array': {
        'factory': schema.List,
        'options': {
            'value_type': schema.TextLine()
        }
    },
    'image': {
        'factory': schema.List,
        'widget': ImageRelatedItemFieldWidget,
        'options': {
            'constraint': validate_image,
            'value_type': schema.Choice(
                vocabulary='plone.app.vocabularies.Catalog'
            )
        }
    },
    'images': {
        'factory': schema.List,
        'widget': ImageRelatedItemsFieldWidget,
        'options': {
            'value_type': schema.Choice(
                vocabulary='plone.app.vocabularies.Catalog'
            )
        }
    },
    'resources': {
        'factory': schema.List,
        'widget': RelatedItemsFieldWidget,
        'options': {
            'constraint': validate_content,
            'value_type': schema.Choice(
                vocabulary='plone.app.vocabularies.Catalog'
            )
        }
    },
    'resource': {
        'factory': schema.List,
        'widget': RelatedItemFieldWidget,
        'options': {
            'constraint': validate_content,
            'value_type': schema.Choice(
                vocabulary='plone.app.vocabularies.Catalog'
            )
        }
    }
}


class DynamicTile(BaseTile):
    implements(IPersistentTile)

    def get_object(self, val):
        if isinstance(val, list):
            if len(val) == 0:
                return None
            val = val[0]
        return self.utils.get_object(val)

    def render(self):
        mng = get_tile_manager(self.request)
        if self.data.get('tile_id') is None:
            logger.info('Can not render tile, no tile_id: {}'.format(
                self.data
            ))
            return '<div></div>'
        template = mng.get_template(self.data['tile_id'])
        site = api.portal.get()
        site_url = site.absolute_url()
        registry = getUtility(IRegistry)
        public_url = registry.get('plone.public_url', None)
        if not public_url:
            public_url = site_url
            if not api.user.is_anonymous():
                site_url = public_url

        boundNames = {
            'context': self.context,
            'request': self.request,
            'view': self,
            'portal_url': site_url,
            'public_url': public_url,
            'site_url': site_url,
            'registry': registry,
            'portal': site,
            'utils': self.utils,
            'data': self.data,
            'get_object': self.get_object
        }
        zpt = template.__of__(self.context)
        return zpt._exec(boundNames, [], {})


class IDynamicTileSchema(model.Schema):

    directives.mode(tile_id='hidden')
    tile_id = schema.TextLine(
        title=u'Selected dynamic tile',
        required=True,
        default=None)


class TileManager(object):

    def __init__(self):
        pass

    def get_tile_directory(self):
        theme = getCurrentTheme()
        if theme:
            directory = queryResourceDirectory(
                THEME_RESOURCE_NAME, theme)
            return directory['tiles']
        else:
            raise NotFound

    def read_file(self, fi):
        if fi.__class__.__name__ == "FilesystemFile":
            data = IRawReadFile(fi).read()
        else:
            data = str(fi.data)
        return data

    def read_tile_config(self, directory):
        data = self.read_file(directory['config.json'])
        try:
            return json.loads(data)
        except Exception:
            logger.info('Could not parse tile config {}: {}'.format(
                directory.__name__, data
            ), exc_info=True)
            raise NotFound

    @forever.memoize
    def get_tile(self, tile_id):
        tiles_directory = self.get_tile_directory()
        tile_folder = tiles_directory[tile_id]
        return self.read_tile_config(tile_folder)

    @forever.memoize
    def get_tile_fields(self, tile_id):
        fields = OrderedDict()
        for field in self.get_tile(tile_id)['fields']:
            field = field.copy()
            try:
                factory = FIELD_TYPE_MAPPING[field.pop('type', 'text')]
            except KeyError:
                continue
            for name in ('title', 'description'):
                if name in field:
                    field[name] = unicode(field[name])
            for name in list(field.keys())[:]:
                if name not in ('title', 'description', 'required', 'vocabulary',
                                'default', 'name'):
                    del field[name]

            if 'required' not in field:
                field['required'] = True

            try:
                field_name = str(field.pop('name'))
                fields[field_name] = (factory, field)
            except Exception:
                logger.info('Could not create field on tile {}, with options {}'.format(
                    tile_id, field
                ), exc_info=True)
        return fields

    @forever.memoize
    def get_schema(self, tile_id):
        widget_tags = {}
        fields = OrderedDict()
        for field_name, (factory, data) in self.get_tile_fields(tile_id).items():
            if isinstance(factory, dict):
                if 'widget' in factory:
                    widget_tags[field_name] = factory['widget']
                if 'options' in factory:
                    data.update(factory['options'])
                factory = factory['factory']
            fields[field_name] = factory(**data)

        schema = SchemaClass(
            name=tile_id,
            bases=(IDynamicTileSchema,),
            # we're mimicking plone.supermodel here so let's us
            # same module
            __module__='plone.supermodel.generated',
            attrs=fields
        )
        schema.setTaggedValue(WIDGETS_KEY, widget_tags)
        return schema

    @forever.memoize
    def get_template(self, tile_id):
        tiles_directory = self.get_tile_directory()
        tile_folder = tiles_directory[tile_id]
        fi = tile_folder['template.html']
        if fi.__class__.__name__ == "FilesystemFile":
            data = IRawReadFile(fi).read()
        else:
            data = str(fi.data)
        return ZopePageTemplate(tile_id, text=data)

    @forever.memoize
    def get_tiles(self):
        try:
            tiles_directory = self.get_tile_directory()
        except NotFound:
            return []
        tiles = []
        for tile_id in tiles_directory.listDirectory():
            directory = tiles_directory[tile_id]
            try:
                config = self.read_tile_config(directory)
            except NotFound:
                continue
            tiles.append({
                'id': tile_id,
                'category': config.get('category', 'app'),
                'title': config.get('title', 'Tile'),
                'description': config.get('description', ''),
                'name': tile_id,
                'weight': config.get('weight', 200),
                'hidden': config.get('hidden', False)
            })
        return tiles


_cache = {}


def get_tile_manager(request=None):
    if request is None:
        request = getRequest()
    cache_key = '{}.manager'.format(CACHE_KEY)
    if cache_key not in request.environ:
        if api.env.debug_mode():
            request.environ[cache_key] = TileManager()
        else:
            # check for value in cache
            theme = getCurrentTheme()
            item_cache_key = '{}.manager.{}'.format(CACHE_KEY, theme)
            if item_cache_key not in _cache:
                _cache[item_cache_key] = {
                    'when': time.time(),
                    'value': TileManager()
                }
            else:
                registry = queryUtility(IRegistry)
                theme_cache_time = getattr(registry, '_theme_cache_mtime', 0)
                if theme_cache_time > _cache[item_cache_key]['when']:
                    _cache[item_cache_key] = {
                        'when': time.time(),
                        'value': TileManager()
                    }
            request.environ[cache_key] = _cache[item_cache_key]['value']
    return request.environ[cache_key]


class AddForm(add.DefaultAddForm):

    @property
    def schema(self):
        mng = get_tile_manager(self.request)
        return mng.get_schema(self.tile_id)

    @property
    def label(self):
        mng = get_tile_manager(self.request)
        tile = mng.get_tile(self.tile_id)
        return 'Add ' + tile.get('title', 'Tile')

    @property
    def description(self):
        mng = get_tile_manager(self.request)
        tile = mng.get_tile(self.tile_id)
        return tile.get('description', '')

    def update(self):
        super(AddForm, self).update()
        if not self.widgets['tile_id'].value:
            self.widgets['tile_id'].value = unicode(
                self.widgets['tile_id'].extract() or self.tile_id)


class AddView(add.DefaultAddView):
    form = AddForm

    def __init__(self, context, request, tileType):
        super(AddView, self).__init__(context, request, tileType)
        self.form_instance.tile_id = unicode(self.request.get(
            'castle.cms.dynamic.tile_id', ''))
        # weird bug where it isn't passing in as unicode?
        self.request['castle.cms.dynamic.tile_id'] = self.form_instance.tile_id


class AddTileTraversal(traversal.AddTile):
    def __call__(self):
        self.errors = {}
        self.request['disable_border'] = True
        if 'form.button.Create' in self.request:
            newTileType = self.request.get('tiletype', None)
            if newTileType is None:
                self.errors['tiletype'] = _(u"You must select the type of " +
                                            u"tile to create")

            if len(self.errors) == 0:
                extra = ''
                if newTileType == 'castle.cms.dynamic':
                    extra = '?castle.cms.dynamic.tile_id={}'.format(
                        self.request.get('tile_id', ''))
                self.request.response.redirect("%s/@@add-tile/%s%s" % (
                    self.context.absolute_url(), newTileType, extra))
                return ''

        return self.index()


class EditForm(edit.DefaultEditForm):

    @property
    def label(self):
        mng = get_tile_manager(self.request)
        tile = mng.get_tile(self.tile_id)
        return 'Edit ' + tile.get('title', 'Tile')

    @property
    def description(self):
        mng = get_tile_manager(self.request)
        tile = mng.get_tile(self.tile_id)
        return tile.get('description', '')

    @property
    def schema(self):
        mng = get_tile_manager(self.request)
        data = self.getContent()
        return mng.get_schema(data['tile_id'])


class EditView(edit.DefaultEditView):
    form = EditForm



from AccessControl.SecurityManagement import newSecurityManager
from castle.cms.utils import retriable
from castle.cms.constants import TRASH_LOG_KEY
from DateTime import DateTime
from plone import api
from plone.app.linkintegrity.exceptions import LinkIntegrityNotificationException
from plone.locking.interfaces import ILockable
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from tendo import singleton
from zope.annotation.interfaces import IAnnotations
from zope.component.hooks import setSite

import transaction
import logging


logger = logging.getLogger('castle.cms')


@retriable(sync=True)
def empty(site):
    setSite(site)
    catalog = api.portal.get_tool('portal_catalog')
    end = DateTime() - 30
    query = dict(
        modified={
            'query': (DateTime('1999/09/09'), end),
            'range': 'min:max'
        },
        trashed=True)
    deleted_count = 0
    empty_trash_log = ''
    trash_brains = catalog(**query)
    for brain in trash_brains:
        ob = brain.getObject()
        lockable = ILockable(ob, None)
        if lockable and lockable.locked():
            lockable.clear_locks()
        try:
            api.content.delete(ob)
            deleted_count += 1
        except LinkIntegrityNotificationException:
            logger.warn('Did not delete {} from trash, link integrity'.format(ob))
            if not empty_trash_log:
                empty_trash_log = 'The following items could not be deleted\n' \
                    'They may still be linked by other content.\n'
            empty_trash_log += '{}\n'.format(ob.absolute_url())
            pass
        if deleted_count % 20 == 0:
            transaction.commit()
    site_annotations = IAnnotations(site)

    empty_trash_log += 'Found {} items recycled 30+ days ago, deleted {} items'.format(
        len(trash_brains),
        deleted_count
    )
    site_annotations[TRASH_LOG_KEY] = empty_trash_log
    transaction.commit()


def run(app):
    singleton.SingleInstance('emptytrash')

    user = app.acl_users.getUser('admin')  # noqa
    newSecurityManager(None, user.__of__(app.acl_users))  # noqa

    for oid in app.objectIds():  # noqa
        obj = app[oid]  # noqa
        if IPloneSiteRoot.providedBy(obj):
            empty(obj)


if __name__ == '__main__':
    run(app)  # noqa
